# pylint: disable=no-name-in-module
from __future__ import annotations

import os
import weakref
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Dict, Literal

from bec_lib.device import ReadoutPriority
from bec_lib.logger import bec_logger
from bec_qthemes._icon.material_icons import material_icon
from qtpy.QtCore import QSize, Qt, QTimer
from qtpy.QtGui import QAction, QColor, QIcon  # type: ignore
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QMenu,
    QSizePolicy,
    QStyledItemDelegate,
    QToolBar,
    QToolButton,
    QWidget,
)

import bec_widgets
from bec_widgets.utils.toolbars.splitter import ResizableSpacer
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import BECDeviceFilter
from bec_widgets.widgets.control.device_input.device_combobox.device_combobox import DeviceComboBox

logger = bec_logger.logger

MODULE_PATH = os.path.dirname(bec_widgets.__file__)


def create_action_with_text(toolbar_action, toolbar: QToolBar, min_size: QSize | None = None):
    """
    Helper function to create a toolbar button with text beside or under the icon.

    Args:
        toolbar_action(ToolBarAction): The toolbar action to create the button for.
        toolbar(ModularToolBar): The toolbar to add the button to.
        min_size(QSize, optional): The minimum size for the button. Defaults to None.
    """

    btn = QToolButton(parent=toolbar)
    if min_size is not None:
        btn.setMinimumSize(min_size)
    if getattr(toolbar_action, "label_text", None):
        toolbar_action.action.setText(toolbar_action.label_text)
    if getattr(toolbar_action, "tooltip", None):
        toolbar_action.action.setToolTip(toolbar_action.tooltip)
        btn.setToolTip(toolbar_action.tooltip)

    btn.setDefaultAction(toolbar_action.action)
    btn.setAutoRaise(True)
    if toolbar_action.text_position == "beside":
        btn.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
    else:
        btn.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
    btn.setText(toolbar_action.label_text)
    toolbar.addWidget(btn)


class NoCheckDelegate(QStyledItemDelegate):
    """To reduce space in combo boxes by removing the checkmark."""

    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        # Remove any check indicator
        option.checkState = Qt.CheckState.Unchecked


class LongPressToolButton(QToolButton):
    def __init__(self, *args, long_press_threshold=500, **kwargs):
        super().__init__(*args, **kwargs)
        self.long_press_threshold = long_press_threshold
        self._long_press_timer = QTimer(self)
        self._long_press_timer.setSingleShot(True)
        self._long_press_timer.timeout.connect(self.handleLongPress)
        self._pressed = False
        self._longPressed = False

    def mousePressEvent(self, event):
        self._pressed = True
        self._longPressed = False
        self._long_press_timer.start(self.long_press_threshold)
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self._pressed = False
        if self._longPressed:
            self._longPressed = False
            self._long_press_timer.stop()
            event.accept()  # Prevent normal click action after a long press
            return
        self._long_press_timer.stop()
        super().mouseReleaseEvent(event)

    def handleLongPress(self):
        if self._pressed:
            self._longPressed = True
            self.showMenu()


class ToolBarAction(ABC):
    """
    Abstract base class for toolbar actions.

    Args:
        icon_path (str, optional): The name of the icon file from `assets/toolbar_icons`. Defaults to None.
        tooltip (str, optional): The tooltip for the action. Defaults to None.
        checkable (bool, optional): Whether the action is checkable. Defaults to False.
    """

    def __init__(
        self, icon_path: str | None = None, tooltip: str | None = None, checkable: bool = False
    ):
        self.icon_path = (
            os.path.join(MODULE_PATH, "assets", "toolbar_icons", icon_path) if icon_path else None
        )
        self.tooltip: str = tooltip or ""
        self.checkable: bool = checkable
        self.action: QAction

    @abstractmethod
    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        """Adds an action or widget to a toolbar.

        Args:
            toolbar (QToolBar): The toolbar to add the action or widget to.
            target (QWidget): The target widget for the action.
        """

    def cleanup(self):
        """Cleans up the action, if necessary."""
        pass


class IconAction(ToolBarAction):
    @abstractmethod
    def get_icon(self) -> QIcon: ...


class SeparatorAction(ToolBarAction):
    """Separator action for the toolbar."""

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        toolbar.addSeparator()


class QtIconAction(IconAction):
    def __init__(
        self,
        standard_icon,
        tooltip=None,
        checkable=False,
        label_text: str | None = None,
        text_position: Literal["beside", "under"] | None = None,
        parent=None,
    ):
        """
        Action with a standard Qt icon for the toolbar.

        Args:
            standard_icon: The standard icon from QStyle.
            tooltip(str, optional): The tooltip for the action. Defaults to None.
            checkable(bool, optional): Whether the action is checkable. Defaults to False.
            label_text(str | None, optional): Optional label text to display beside or under the icon.
            text_position(Literal["beside", "under"] | None, optional): Position of text relative to icon.
            parent(QWidget or None, optional): Parent widget for the underlying QAction.
        """
        super().__init__(icon_path=None, tooltip=tooltip, checkable=checkable)
        self.standard_icon = standard_icon
        self.icon = QApplication.style().standardIcon(standard_icon)
        self.action = QAction(icon=self.icon, text=self.tooltip, parent=parent)
        self.action.setCheckable(self.checkable)
        self.label_text = label_text
        self.text_position = text_position

    def add_to_toolbar(self, toolbar, target):
        if self.label_text is not None:
            create_action_with_text(toolbar_action=self, toolbar=toolbar)
        else:
            toolbar.addAction(self.action)

    def get_icon(self):
        return self.icon


class MaterialIconAction(IconAction):
    """
    Action with a Material icon for the toolbar.

    Args:
        icon_name (str): The name of the Material icon.
        tooltip (str): The tooltip for the action.
        checkable (bool, optional): Whether the action is checkable. Defaults to False.
        filled (bool, optional): Whether the icon is filled. Defaults to False.
        color (str | tuple | QColor | dict[Literal["dark", "light"], str] | None, optional): The color of the icon.
            Defaults to None.
        label_text (str | None, optional): Optional label text to display beside or under the icon.
        text_position (Literal["beside", "under"] | None, optional): Position of text relative to icon.
        parent (QWidget or None, optional): Parent widget for the underlying QAction.
    """

    def __init__(
        self,
        icon_name: str,
        tooltip: str,
        *,
        checkable: bool = False,
        filled: bool = False,
        color: str | tuple | QColor | dict[Literal["dark", "light"], str] | None = None,
        label_text: str | None = None,
        text_position: Literal["beside", "under"] | None = None,
        parent=None,
    ):
        """
        MaterialIconAction for toolbar: if label_text and text_position are provided, show text beside or under icon.
        This enables per-action icon text without breaking the existing API.
        """
        super().__init__(icon_path=None, tooltip=tooltip, checkable=checkable)
        self.icon_name = icon_name
        self.filled = filled
        self.color = color
        self.label_text = label_text
        self.text_position = text_position
        # Generate the icon using the material_icon helper
        self.icon: QIcon = material_icon(
            self.icon_name,
            size=(20, 20),
            convert_to_pixmap=False,
            filled=self.filled,
            color=self.color,
        )  # type: ignore
        if parent is None:
            logger.warning(
                "MaterialIconAction was created without a parent. Please consider adding one. Using None as parent may cause issues."
            )
        self.action = QAction(icon=self.icon, text=self.tooltip, parent=parent)
        self.action.setCheckable(self.checkable)

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        """
        Adds the action to the toolbar.

        Args:
            toolbar(QToolBar): The toolbar to add the action to.
            target(QWidget): The target widget for the action.
        """
        if self.label_text is not None:
            create_action_with_text(toolbar_action=self, toolbar=toolbar)
        else:
            toolbar.addAction(self.action)

    def get_icon(self):
        """
        Returns the icon for the action.

        Returns:
            QIcon: The icon for the action.
        """
        return self.icon


class DeviceSelectionAction(ToolBarAction):
    """
    Action for selecting a device in a combobox.

    Args:
        device_combobox (DeviceComboBox): The combobox for selecting the device.
        label (str): The label for the combobox.
    """

    def __init__(self, /, device_combobox: DeviceComboBox, label: str | None = None):
        super().__init__()
        self.label = label
        self.device_combobox = device_combobox
        self.device_combobox.currentIndexChanged.connect(lambda: self.set_combobox_style("#ffa700"))

    def add_to_toolbar(self, toolbar, target):
        widget = QWidget(parent=target)
        layout = QHBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        if self.label is not None:
            label = QLabel(text=f"{self.label}", parent=target)
            layout.addWidget(label)
        if self.device_combobox is not None:
            layout.addWidget(self.device_combobox)
            toolbar.addWidget(widget)

    def set_combobox_style(self, color: str):
        self.device_combobox.setStyleSheet(f"QComboBox {{ background-color: {color}; }}")


class SwitchableToolBarAction(IconAction):
    """
    A split toolbar action that combines a main action and a drop-down menu for additional actions.

    The main button displays the currently selected action's icon and tooltip. Clicking on the main button
    triggers that action. Clicking on the drop-down arrow displays a menu with alternative actions. When an
    alternative action is selected, it becomes the new default and its callback is immediately executed.

    This design mimics the behavior seen in Adobe Photoshop or Affinity Designer toolbars.

    Args:
        actions (dict): A dictionary mapping a unique key to a ToolBarAction instance.
        initial_action (str, optional): The key of the initial default action. If not provided, the first action is used.
        tooltip (str, optional): An optional tooltip for the split action; if provided, it overrides the default action's tooltip.
        checkable (bool, optional): Whether the action is checkable. Defaults to True.
        parent (QWidget, optional): Parent widget for the underlying QAction.
    """

    def __init__(
        self,
        actions: Dict[str, IconAction],
        initial_action: str | None = None,
        tooltip: str | None = None,
        checkable: bool = True,
        default_state_checked: bool = False,
        parent=None,
    ):
        super().__init__(icon_path=None, tooltip=tooltip, checkable=checkable)
        self.actions = actions
        self.current_key = initial_action if initial_action is not None else next(iter(actions))
        self.parent = parent
        self.checkable = checkable
        self.default_state_checked = default_state_checked
        self.main_button = None
        self.menu_actions: Dict[str, QAction] = {}

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        """
        Adds the split action to the toolbar.

        Args:
            toolbar (QToolBar): The toolbar to add the action to.
            target (QWidget): The target widget for the action.
        """
        self.main_button = LongPressToolButton(toolbar)
        self.main_button.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
        self.main_button.setCheckable(self.checkable)
        default_action = self.actions[self.current_key]
        self.main_button.setIcon(default_action.get_icon())
        self.main_button.setToolTip(default_action.tooltip or "")
        self.main_button.clicked.connect(self._trigger_current_action)
        menu = QMenu(self.main_button)
        for key, action_obj in self.actions.items():
            menu_action = QAction(
                icon=action_obj.get_icon(), text=action_obj.tooltip, parent=self.main_button
            )
            menu_action.setIconVisibleInMenu(True)
            menu_action.setCheckable(self.checkable)
            menu_action.setChecked(key == self.current_key)
            menu_action.triggered.connect(lambda checked, k=key: self.set_default_action(k))
            menu.addAction(menu_action)
        self.main_button.setMenu(menu)
        if self.default_state_checked:
            self.main_button.setChecked(True)
        self.action = toolbar.addWidget(self.main_button)

    def _trigger_current_action(self):
        """
        Triggers the current action associated with the main button.
        """
        action_obj = self.actions[self.current_key]
        action_obj.action.trigger()

    def set_default_action(self, key: str):
        """
        Sets the default action for the split action.

        Args:
            key(str): The key of the action to set as default.
        """
        if self.main_button is None:
            return
        self.current_key = key
        new_action = self.actions[self.current_key]
        self.main_button.setIcon(new_action.get_icon())
        self.main_button.setToolTip(new_action.tooltip)
        # Update check state of menu items
        for k, menu_act in self.actions.items():
            menu_act.action.setChecked(False)
        new_action.action.trigger()
        # Active action chosen from menu is always checked, uncheck through main button
        if self.checkable:
            new_action.action.setChecked(True)
            self.main_button.setChecked(True)

    def block_all_signals(self, block: bool = True):
        """
        Blocks or unblocks all signals for the actions in the toolbar.

        Args:
            block (bool): Whether to block signals. Defaults to True.
        """
        if self.main_button is not None:
            self.main_button.blockSignals(block)

        for action in self.actions.values():
            action.action.blockSignals(block)

    @contextmanager
    def signal_blocker(self):
        """
        Context manager to block signals for all actions in the toolbar.
        """
        self.block_all_signals(True)
        try:
            yield
        finally:
            self.block_all_signals(False)

    def set_state_all(self, state: bool):
        """
        Uncheck all actions in the toolbar.
        """
        for action in self.actions.values():
            action.action.setChecked(state)
        if self.main_button is None:
            return
        self.main_button.setChecked(state)

    def get_icon(self) -> QIcon:
        return self.actions[self.current_key].get_icon()


class WidgetAction(ToolBarAction):
    """
    Action for adding any widget to the toolbar.
    Please note that the injected widget must be life-cycled by the parent widget,
    i.e., the widget must be properly cleaned up outside of this action. The WidgetAction
    will not perform any cleanup on the widget itself, only on the container that holds it.

    Args:
        label (str|None): The label for the widget.
        widget (QWidget): The widget to be added to the toolbar.
        adjust_size (bool): Whether to adjust the size of the widget based on its contents. Defaults to True.
    """

    def __init__(
        self, *, widget: QWidget, label: str | None = None, adjust_size: bool = True, parent=None
    ):
        super().__init__(icon_path=None, tooltip=label, checkable=False)
        self.label = label
        self.widget = widget
        self.container = None
        self.adjust_size = adjust_size

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        """
        Adds the widget to the toolbar.

        Args:
            toolbar (QToolBar): The toolbar to add the widget to.
            target (QWidget): The target widget for the action.
        """
        self.container = QWidget(parent=target)
        layout = QHBoxLayout(self.container)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        if self.label is not None:
            label_widget = QLabel(text=f"{self.label}", parent=target)
            label_widget.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
            label_widget.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignRight)
            layout.addWidget(label_widget)

        if isinstance(self.widget, QComboBox) and self.adjust_size:
            self.widget.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToContents)

            size_policy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
            self.widget.setSizePolicy(size_policy)

            self.widget.setMinimumWidth(self.calculate_minimum_width(self.widget))

        layout.addWidget(self.widget)

        toolbar.addWidget(self.container)
        # Store the container as the action to allow toggling visibility.
        self.action = self.container  # type: ignore

    def cleanup(self):
        """
        Cleans up the action by closing and deleting the container widget.
        This method will be called automatically when the toolbar is cleaned up.
        """
        if self.container is not None:
            self.container.close()
            self.container.deleteLater()
        return super().cleanup()

    @staticmethod
    def calculate_minimum_width(combo_box: QComboBox) -> int:
        font_metrics = combo_box.fontMetrics()
        max_width = max(font_metrics.width(combo_box.itemText(i)) for i in range(combo_box.count()))  # type: ignore
        return max_width + 60


class SplitterAction(ToolBarAction):
    """
    Action for adding a draggable splitter/spacer to the toolbar.

    This creates a resizable spacer that allows users to control how much space
    is allocated to toolbar sections before and after it. When dragged, it expands/contracts,
    pushing other toolbar elements left or right.

    Args:
        orientation (Literal["horizontal", "vertical", "auto"]): The orientation of the splitter.
        parent (QWidget): The parent widget.
        initial_width (int): Fixed size of the spacer in pixels along the toolbar's orientation (default: 20).
        min_width (int | None): Minimum size of the target widget along the orientation axis (width for horizontal, height for vertical). If ``None``, no minimum constraint is applied.
        max_width (int | None): Maximum size of the target widget along the orientation axis (width for horizontal, height for vertical). If ``None``, no maximum constraint is applied.
        target_widget (QWidget | None): Widget whose size (width or height, depending on orientation) is controlled by the spacer within the given min/max bounds.
    """

    def __init__(
        self,
        orientation: Literal["horizontal", "vertical", "auto"] = "auto",
        parent=None,
        initial_width=20,
        min_width: int | None = None,
        max_width: int | None = None,
        target_widget=None,
    ):
        super().__init__(icon_path=None, tooltip="Drag to resize toolbar sections", checkable=False)
        self.orientation = orientation
        self.initial_width = initial_width
        self.min_width = min_width
        self.max_width = max_width
        self._splitter_widget = None
        self._target_widget = target_widget

    def _resolve_orientation(self, toolbar: QToolBar) -> Literal["horizontal", "vertical"]:
        if self.orientation in (None, "auto"):
            return (
                "horizontal" if toolbar.orientation() == Qt.Orientation.Horizontal else "vertical"
            )
        return self.orientation

    def set_target_widget(self, widget):
        """Set the target widget after creation."""
        self._target_widget = widget
        if self._splitter_widget:
            self._splitter_widget.set_target_widget(widget)

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        """
        Adds the splitter/spacer to the toolbar.

        Args:
            toolbar (QToolBar): The toolbar to add the splitter to.
            target (QWidget): The target widget for the action.
        """

        effective_orientation = self._resolve_orientation(toolbar)
        self._splitter_widget = ResizableSpacer(
            parent=target,
            orientation=effective_orientation,
            initial_width=self.initial_width,
            min_target_size=self.min_width,
            max_target_size=self.max_width,
            target_widget=self._target_widget,
        )
        toolbar.addWidget(self._splitter_widget)
        self.action = self._splitter_widget  # type: ignore

    def cleanup(self):
        """Clean up the splitter widget."""
        if self._splitter_widget is not None:
            self._splitter_widget.close()
            self._splitter_widget.deleteLater()
        return super().cleanup()


class ExpandableMenuAction(ToolBarAction):
    """
    Action for an expandable menu in the toolbar.

    Args:
        label (str): The label for the menu.
        actions (dict): A dictionary of actions to populate the menu.
        icon_path (str, optional): The path to the icon file. Defaults to None.
    """

    def __init__(self, label: str, actions: dict[str, IconAction], icon_path: str | None = None):
        super().__init__(icon_path, label)
        self.actions = actions
        self._button_ref: weakref.ReferenceType[QToolButton] | None = None
        self._menu_ref: weakref.ReferenceType[QMenu] | None = None

    def add_to_toolbar(self, toolbar: QToolBar, target: QWidget):
        button = QToolButton(toolbar)
        button.setObjectName("toolbarMenuButton")
        button.setAutoRaise(True)
        if self.icon_path:
            button.setIcon(QIcon(self.icon_path))
        button.setText(self.tooltip)
        button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        button.setStyleSheet("""
                   QToolButton {
                       font-size: 14px;
                   }
                   QMenu {
                       font-size: 14px;
                   }
               """)
        menu = QMenu(button)
        for action_container in self.actions.values():
            action: QAction = action_container.action
            action.setIconVisibleInMenu(True)
            if action_container.icon_path:
                icon = QIcon()
                icon.addFile(action_container.icon_path, size=QSize(20, 20))
                action.setIcon(icon)
            elif hasattr(action, "get_icon") and callable(action_container.get_icon):
                sub_icon = action_container.get_icon()
                if sub_icon and not sub_icon.isNull():
                    action.setIcon(sub_icon)
            action.setCheckable(action_container.checkable)
            menu.addAction(action)
        button.setMenu(menu)
        toolbar.addWidget(button)
        self._button_ref = weakref.ref(button)
        self._menu_ref = weakref.ref(menu)

    def get_toolbar_button(self) -> QToolButton | None:
        return self._button_ref() if self._button_ref else None

    def get_menu(self) -> QMenu | None:
        return self._menu_ref() if self._menu_ref else None


class DeviceComboBoxAction(WidgetAction):
    """
    Action for a device selection combobox in the toolbar.

    Args:
        label (str): The label for the combobox.
        device_combobox (QComboBox): The combobox for selecting the device.
    """

    def __init__(
        self,
        target_widget: QWidget,
        device_filter: list[BECDeviceFilter] | None = None,
        readout_priority_filter: (
            str | ReadoutPriority | list[str] | list[ReadoutPriority] | None
        ) = None,
        tooltip: str | None = None,
        add_empty_item: bool = False,
        no_check_delegate: bool = False,
    ):
        self.combobox = DeviceComboBox(
            parent=target_widget,
            device_filter=device_filter,
            readout_priority_filter=readout_priority_filter,
        )
        super().__init__(widget=self.combobox, adjust_size=False)

        if add_empty_item:
            self.combobox.addItem("", None)
            self.combobox.setCurrentText("")
        if tooltip is not None:
            self.combobox.setToolTip(tooltip)
        if no_check_delegate:
            self.combobox.setItemDelegate(NoCheckDelegate(self.combobox))

    def cleanup(self):
        """
        Cleans up the action by closing and deleting the combobox widget.
        This method will be called automatically when the toolbar is cleaned up.
        """
        if self.combobox is not None:
            self.combobox.close()
            self.combobox.deleteLater()
        return super().cleanup()


class TutorialAction(MaterialIconAction):
    """
    Action for starting a guided tutorial/help tour.

    This action automatically initializes a GuidedTour instance and provides
    methods to register widgets and start tours.

    Args:
        main_window (QWidget): The main window widget for the guided tour overlay.
        tooltip (str, optional): The tooltip for the action. Defaults to "Start Guided Tutorial".
        parent (QWidget or None, optional): Parent widget for the underlying QAction.
    """

    def __init__(self, main_window: QWidget, tooltip: str = "Start Guided Tutorial", parent=None):
        super().__init__(
            icon_name="help",
            tooltip=tooltip,
            checkable=False,
            filled=False,
            color=None,
            parent=parent,
        )

        from bec_widgets.utils.guided_tour import GuidedTour

        self.guided_help = GuidedTour(main_window)
        self.main_window = main_window

        # Connect the action to start the tour
        self.action.triggered.connect(self.start_tour)

    def register_widget(self, widget: QWidget, text: str, widget_name: str = "") -> str:
        """
        Register a widget for the guided tour.

        Args:
            widget (QWidget): The widget to highlight during the tour.
            text (str): The help text to display.
            widget_name (str, optional): Optional name for the widget.

        Returns:
            str: Unique ID for the registered widget.
        """
        return self.guided_help.register_widget(widget=widget, text=text, title=widget_name)

    def start_tour(self):
        """Start the guided tour with all registered widgets."""
        registered_widgets = self.guided_help.get_registered_widgets()
        if registered_widgets:
            # Create tour from all registered widgets
            step_ids = list(registered_widgets.keys())
            if self.guided_help.create_tour(step_ids):
                self.guided_help.start_tour()
            else:
                logger.warning("Failed to create guided tour")
        else:
            logger.warning("No widgets registered for guided tour")

    def has_registered_widgets(self) -> bool:
        """Check if any widgets have been registered for the tour."""
        return len(self.guided_help.get_registered_widgets()) > 0

    def clear_registered_widgets(self):
        """Clear all registered widgets."""
        self.guided_help.clear_registrations()

    def cleanup(self):
        """Clean up the guided help instance."""
        if hasattr(self, "guided_help"):
            self.guided_help.stop_tour()
        super().cleanup()



"""Module with a config view for the device manager."""

from __future__ import annotations

import traceback

import yaml
from bec_lib.logger import bec_logger
from qtpy import QtCore, QtWidgets

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.widgets.editors.monaco.monaco_widget import MonacoWidget

logger = bec_logger.logger


class DMConfigView(QtWidgets.QWidget):
    """Widget to show the config of a selected device in YAML format."""

    RPC = False

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.stacked_layout = QtWidgets.QStackedLayout()
        self.stacked_layout.setContentsMargins(0, 0, 0, 0)
        self.stacked_layout.setSpacing(0)
        self.setLayout(self.stacked_layout)

        # Monaco widget
        self.monaco_editor = MonacoWidget(parent=self)
        self._customize_monaco()
        self.stacked_layout.addWidget(self.monaco_editor)

        # Overlay widget
        self._overlay_text = "Select a single device to view its config."
        self._overlay_widget = QtWidgets.QLabel(text=self._overlay_text)
        self._customize_overlay()
        self.stacked_layout.addWidget(self._overlay_widget)
        self.stacked_layout.setCurrentWidget(self._overlay_widget)

    def _customize_monaco(self):
        """Customize the Monaco editor for YAML display."""
        self.monaco_editor.set_language("yaml")
        self.monaco_editor.set_vim_mode_enabled(False)
        self.monaco_editor.set_minimap_enabled(False)
        self.monaco_editor.set_readonly(True)
        self.monaco_editor.editor.set_scroll_beyond_last_line_enabled(False)
        self.monaco_editor.editor.set_line_numbers_mode("off")

    def _customize_overlay(self):
        """Customize the overlay widget."""
        self._overlay_widget.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self._overlay_widget.setAutoFillBackground(True)
        self._overlay_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding
        )

    @SafeSlot(dict)
    def on_select_config(self, device: list[dict]):
        """
        Handle selection of a device from the device table. If more than one device is selected,
        show an overlay message. Otherwise, display the device config in YAML format.

        Args:
            device (list[dict]): The selected device configuration.
        """
        if len(device) != 1:
            text = ""
            self.stacked_layout.setCurrentWidget(self._overlay_widget)
        else:
            try:
                # Cast set to list to ensure proper YAML dumping
                cfg = device[0]
                for k, v in cfg.items():
                    if isinstance(v, set):
                        cfg[k] = list(v)
                text = yaml.dump(cfg, default_flow_style=False)
                self.stacked_layout.setCurrentWidget(self.monaco_editor)
            except Exception:
                content = traceback.format_exc()
                logger.error(f"Error converting device to YAML:\n{content}")
                text = ""
                self.stacked_layout.setCurrentWidget(self._overlay_widget)
        self.monaco_editor.set_readonly(False)  # Enable editing
        text = text.rstrip()
        self.monaco_editor.set_text(text)
        self.monaco_editor.set_readonly(True)  # Disable editing again


if __name__ == "__main__":  # pragma: no cover
    import sys

    from bec_qthemes import apply_theme
    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    apply_theme("dark")
    widget = QtWidgets.QWidget()
    layout = QtWidgets.QVBoxLayout(widget)
    widget.setLayout(layout)
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)
    config_view = DMConfigView()
    layout.addWidget(config_view)
    combo_box = QtWidgets.QComboBox()
    config = config_view.client.device_manager._get_redis_device_config()
    combo_box.addItems([""] + [f"{v} : {item.get('name', '')}" for v, item in enumerate(config)])

    def on_select(text):
        if text == "":
            config_view.on_select_config([])
        else:
            index = int(text.split(" : ")[0])
            config_view.on_select_config([config[index]])

    combo_box.currentTextChanged.connect(on_select)
    layout.addWidget(combo_box)
    widget.show()
    sys.exit(app.exec_())



from __future__ import annotations

from qtpy.QtCore import Signal
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QLabel


class ClickableLabel(QLabel):
    clicked = Signal()

    def mouseReleaseEvent(self, ev: QMouseEvent) -> None:
        self.clicked.emit()
        return super().mouseReleaseEvent(ev)



from __future__ import annotations

from bec_qthemes import material_icon
from qtpy.QtCore import Property, Qt, Slot
from qtpy.QtWidgets import QApplication, QHBoxLayout, QPushButton, QToolButton, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import apply_theme


class DarkModeButton(BECWidget, QWidget):

    ICON_NAME = "dark_mode"
    PLUGIN = True
    RPC = False

    def __init__(
        self,
        parent: QWidget | None = None,
        client=None,
        gui_id: str | None = None,
        toolbar: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(parent=parent, client=client, gui_id=gui_id, theme_update=True, **kwargs)
        self.setProperty("skip_settings", True)

        self._dark_mode_enabled = False
        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        if toolbar:
            self.mode_button = QToolButton(parent=parent)
        else:
            self.mode_button = QPushButton(parent=parent)

        self.dark_mode_enabled = self._get_qapp_dark_mode_state()
        self.update_mode_button()
        self.mode_button.clicked.connect(self.toggle_dark_mode)
        self.layout.addWidget(self.mode_button)
        self.setLayout(self.layout)
        self.setFixedSize(40, 40)

    @Slot(str)
    def apply_theme(self, theme: str):
        """
        Apply the theme to the widget.

        Args:
            theme(str, optional): The theme to be applied.
        """
        self.dark_mode_enabled = theme == "dark"
        self.update_mode_button()

    def _get_qapp_dark_mode_state(self) -> bool:
        """
        Get the dark mode state from the QApplication.

        Returns:
            bool: True if dark mode is enabled, False otherwise.
        """
        qapp = QApplication.instance()
        if hasattr(qapp, "theme") and qapp.theme.theme == "dark":
            return True

        return False

    @Property(bool)
    def dark_mode_enabled(self) -> bool:
        """
        The dark mode state. If True, dark mode is enabled. If False, light mode is enabled.
        """
        return self._dark_mode_enabled

    @dark_mode_enabled.setter
    def dark_mode_enabled(self, state: bool) -> None:
        self._dark_mode_enabled = state

    @Slot()
    def toggle_dark_mode(self) -> None:
        """
        Toggle the dark mode state. This will change the theme of the entire
        application to dark or light mode.
        """
        self.dark_mode_enabled = not self.dark_mode_enabled
        self.update_mode_button()
        apply_theme("dark" if self.dark_mode_enabled else "light")

    def update_mode_button(self):
        icon = material_icon(
            "light_mode" if self.dark_mode_enabled else "dark_mode",
            size=(20, 20),
            convert_to_pixmap=False,
        )
        self.mode_button.setIcon(icon)
        self.mode_button.setToolTip("Set Light Mode" if self.dark_mode_enabled else "Set Dark Mode")


if __name__ == "__main__":

    app = QApplication([])
    apply_theme("dark")
    w = DarkModeButton()
    w.show()

    app.exec_()



from __future__ import annotations

from typing import Callable, Literal

from qtpy.QtCore import Qt
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from bec_widgets import SafeSlot


class SaveProfileDialog(QDialog):
    """Dialog for saving workspace profiles with quick select option."""

    def __init__(
        self,
        parent: QWidget | None = None,
        current_name: str = "",
        current_profile_name: str = "",
        *,
        name_exists: Callable[[str], bool] | None = None,
        profile_origin: (
            Callable[[str], Literal["module", "plugin", "settings", "unknown"]] | None
        ) = None,
        origin_label: Callable[[str], str | None] | None = None,
        quick_select_checked: bool = False,
    ):
        super().__init__(parent)
        self.setWindowTitle("Save Workspace Profile")
        self.setModal(True)
        self.resize(400, 160)

        self._name_exists = name_exists or (lambda _: False)
        self._profile_origin = profile_origin or (lambda _: "unknown")
        self._origin_label = origin_label or (lambda _: None)
        self._current_profile_name = current_profile_name.strip()
        self._previous_name_before_overwrite = current_name
        self._block_name_signals = False
        self._block_checkbox_signals = False
        self.overwrite_existing = False

        layout = QVBoxLayout(self)

        # Name input
        name_row = QHBoxLayout()
        name_row.addWidget(QLabel("Profile Name:"))
        self.name_edit = QLineEdit(current_name)
        self.name_edit.setPlaceholderText("Enter profile name...")
        name_row.addWidget(self.name_edit)
        layout.addLayout(name_row)

        # Overwrite checkbox
        self.overwrite_checkbox = QCheckBox("Overwrite current profile")
        self.overwrite_checkbox.setEnabled(bool(self._current_profile_name))
        self.overwrite_checkbox.toggled.connect(self._on_overwrite_toggled)
        layout.addWidget(self.overwrite_checkbox)

        # Quick-select checkbox
        self.quick_select_checkbox = QCheckBox("Include in quick selection.")
        self.quick_select_checkbox.setChecked(quick_select_checked)
        layout.addWidget(self.quick_select_checkbox)

        # Buttons
        btn_row = QHBoxLayout()
        btn_row.addStretch(1)
        self.save_btn = QPushButton("Save")
        self.save_btn.setDefault(True)
        cancel_btn = QPushButton("Cancel")
        self.save_btn.clicked.connect(self.accept)
        cancel_btn.clicked.connect(self.reject)
        btn_row.addWidget(self.save_btn)
        btn_row.addWidget(cancel_btn)
        layout.addLayout(btn_row)

        # Enable/disable save button based on name input
        self.name_edit.textChanged.connect(self._on_name_changed)
        self._update_save_button()

    @SafeSlot(bool)
    def _on_overwrite_toggled(self, checked: bool):
        if self._block_checkbox_signals:
            return
        if not self._current_profile_name:
            return

        self._block_name_signals = True
        if checked:
            self._previous_name_before_overwrite = self.name_edit.text()
            self.name_edit.setText(self._current_profile_name)
            self.name_edit.selectAll()
        else:
            if self.name_edit.text().strip() == self._current_profile_name:
                self.name_edit.setText(self._previous_name_before_overwrite or "")
        self._block_name_signals = False
        self._update_save_button()

    @SafeSlot(str)
    def _on_name_changed(self, _: str):
        if self._block_name_signals:
            return
        text = self.name_edit.text().strip()
        if self.overwrite_checkbox.isChecked() and text != self._current_profile_name:
            self._block_checkbox_signals = True
            self.overwrite_checkbox.setChecked(False)
            self._block_checkbox_signals = False
        self._update_save_button()

    def _update_save_button(self):
        """Enable save button only when name is not empty."""
        self.save_btn.setEnabled(bool(self.name_edit.text().strip()))

    def get_profile_name(self) -> str:
        """Return the entered profile name."""
        return self.name_edit.text().strip()

    def is_quick_select(self) -> bool:
        """Return whether the profile should appear in quick select."""
        return self.quick_select_checkbox.isChecked()

    def _generate_unique_name(self, base: str) -> str:
        candidate_base = base.strip() or "profile"
        suffix = "_custom"
        candidate = f"{candidate_base}{suffix}"
        counter = 1
        while self._name_exists(candidate) or self._profile_origin(candidate) != "unknown":
            candidate = f"{candidate_base}{suffix}_{counter}"
            counter += 1
        return candidate

    def accept(self):
        name = self.get_profile_name()
        if not name:
            return

        self.overwrite_existing = False
        origin = self._profile_origin(name)
        if origin in {"module", "plugin"}:
            source_label = self._origin_label(name)
            if origin == "module":
                provider = source_label or "BEC Widgets"
            else:
                provider = (
                    f"the {source_label} plugin repository"
                    if source_label
                    else "the plugin repository"
                )
            QMessageBox.information(
                self,
                "Read-only profile",
                (
                    f"'{name}' is a default profile provided by {provider} and cannot be overwritten.\n"
                    "Please choose a different name."
                ),
            )
            suggestion = self._generate_unique_name(name)
            self._block_name_signals = True
            self.name_edit.setText(suggestion)
            self.name_edit.selectAll()
            self._block_name_signals = False
            self._block_checkbox_signals = True
            self.overwrite_checkbox.setChecked(False)
            self._block_checkbox_signals = False
            return
        if origin == "settings":
            reply = QMessageBox.question(
                self,
                "Overwrite profile",
                (
                    f"A profile named '{name}' already exists.\n\n"
                    "Overwriting will update both the saved profile and its restore default.\n"
                    "Do you want to continue?"
                ),
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if reply != QMessageBox.StandardButton.Yes:
                suggestion = self._generate_unique_name(name)
                self._block_name_signals = True
                self.name_edit.setText(suggestion)
                self.name_edit.selectAll()
                self._block_name_signals = False
                self._block_checkbox_signals = True
                self.overwrite_checkbox.setChecked(False)
                self._block_checkbox_signals = False
                return
            self.overwrite_existing = True

        super().accept()


class PreviewPanel(QGroupBox):
    """Resizable preview pane that scales its pixmap with aspect ratio preserved."""

    def __init__(self, title: str, pixmap: QPixmap | None, parent: QWidget | None = None):
        super().__init__(title, parent)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self._original: QPixmap | None = pixmap if (pixmap and not pixmap.isNull()) else None

        layout = QVBoxLayout(self)

        self.image_label = QLabel()
        self.image_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.image_label.setMinimumSize(360, 240)
        self.image_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        layout.addWidget(self.image_label, 1)

        if self._original:
            self._update_scaled_pixmap()
        else:
            self.image_label.setText("No preview available")
            self.image_label.setStyleSheet(
                self.image_label.styleSheet() + "color: rgba(255,255,255,0.6); font-style: italic;"
            )

    def setPixmap(self, pixmap: QPixmap | None):
        """
        Set the pixmap to display in the preview panel.

        Args:
            pixmap(QPixmap | None): The pixmap to display. If None or null, clears the preview.

        """
        self._original = pixmap if (pixmap and not pixmap.isNull()) else None
        if self._original:
            self.image_label.setText("")
            self._update_scaled_pixmap()
        else:
            self.image_label.setPixmap(QPixmap())
            self.image_label.setText("No preview available")

    def resizeEvent(self, event):
        super().resizeEvent(event)
        if self._original:
            self._update_scaled_pixmap()

    def _update_scaled_pixmap(self):
        if not self._original:
            return
        size = self.image_label.size()
        if size.width() <= 0 or size.height() <= 0:
            return
        scaled = self._original.scaled(size, Qt.KeepAspectRatio, Qt.SmoothTransformation)
        self.image_label.setPixmap(scaled)


class RestoreProfileDialog(QDialog):
    """
    Confirmation dialog that previews the current profile screenshot against the default baseline.
    """

    def __init__(
        self, parent: QWidget | None, current_pixmap: QPixmap | None, default_pixmap: QPixmap | None
    ):
        super().__init__(parent)
        self.setWindowTitle("Restore Profile to Default")
        self.setModal(True)
        self.resize(880, 480)

        layout = QVBoxLayout(self)

        info_label = QLabel(
            "Restoring will discard your custom layout and replace it with the default profile."
        )
        info_label.setWordWrap(True)
        layout.addWidget(info_label)

        preview_row = QHBoxLayout()
        layout.addLayout(preview_row)

        current_preview = PreviewPanel("Current", current_pixmap, self)
        default_preview = PreviewPanel("Default", default_pixmap, self)

        # Equal expansion left/right
        preview_row.addWidget(current_preview, 1)

        arrow_label = QLabel("\u2192")
        arrow_label.setAlignment(Qt.AlignCenter)
        arrow_label.setStyleSheet("font-size: 32px; padding: 0 16px;")
        arrow_label.setMinimumWidth(40)
        arrow_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        preview_row.addWidget(arrow_label)

        preview_row.addWidget(default_preview, 1)

        # Enforce equal stretch for both previews
        preview_row.setStretch(0, 1)
        preview_row.setStretch(1, 0)
        preview_row.setStretch(2, 1)

        warn_label = QLabel(
            "This action cannot be undone. Do you want to restore the default layout now?"
        )
        warn_label.setWordWrap(True)
        layout.addWidget(warn_label)

        btn_row = QHBoxLayout()
        btn_row.addStretch(1)
        restore_btn = QPushButton("Restore")
        restore_btn.setDefault(True)
        cancel_btn = QPushButton("Cancel")
        restore_btn.clicked.connect(self.accept)
        cancel_btn.clicked.connect(self.reject)
        btn_row.addWidget(restore_btn)
        btn_row.addWidget(cancel_btn)
        layout.addLayout(btn_row)

        # Make the previews take most of the vertical space on resize
        layout.setStretch(0, 0)  # info label
        layout.setStretch(1, 1)  # preview row
        layout.setStretch(2, 0)  # warning label
        layout.setStretch(3, 0)  # buttons

    @staticmethod
    def confirm(
        parent: QWidget | None, current_pixmap: QPixmap | None, default_pixmap: QPixmap | None
    ) -> bool:
        dialog = RestoreProfileDialog(parent, current_pixmap, default_pixmap)
        return dialog.exec() == QDialog.Accepted



import traceback
from pathlib import Path
from typing import Annotated

import copier
import typer
from bec_lib.logger import bec_logger
from bec_lib.plugin_helper import plugin_repo_path
from bec_lib.utils.plugin_manager._constants import ANSWER_KEYS
from bec_lib.utils.plugin_manager._util import existing_data, git_stage_files, make_commit

from bec_widgets.utils.bec_plugin_manager.edit_ui import open_and_watch_ui_editor

logger = bec_logger.logger
_app = typer.Typer(rich_markup_mode="rich")


def _commit_added_widget(repo: Path, name: str):
    git_stage_files(repo, [".copier-answers.yml"])
    git_stage_files(repo / repo.name / "bec_widgets" / "widgets" / name, [])
    make_commit(repo, f"plugin-manager added new widget: {name}")
    logger.info(f"Committing new widget {name}")


def _widget_exists(widget_list: list[dict[str, str | bool]], name: str):
    return name in [w["name"] for w in widget_list]


def _editor_cb(ctx: typer.Context, value: bool):
    if value and not ctx.params["use_ui"]:
        raise typer.BadParameter("Can only open the editor if creating a .ui file!")
    return value


_bold_blue = "\033[34m\033[1m"
_off = "\033[0m"
_USE_UI_MSG = "Generate a .ui file for use in bec-designer."
_OPEN_DESIGNER_MSG = f"""This app can watch for changes and recompile them to a python file imported to the widget whenever it is saved.
To open this editor independently, you can use {_bold_blue}bec-plugin-manager edit-ui [widget_name]{_off}.
Open the created widget .ui file in bec-designer now?"""


@_app.command()
def widget(
    name: Annotated[str, typer.Argument(help="Enter a name for your widget in snake_case")],
    use_ui: Annotated[bool, typer.Option(prompt=_USE_UI_MSG, help=_USE_UI_MSG)] = True,
    open_editor: Annotated[
        bool, typer.Option(prompt=_OPEN_DESIGNER_MSG, help=_OPEN_DESIGNER_MSG, callback=_editor_cb)
    ] = True,
):
    """Create a new widget plugin with the given name.

If [bold white]use_ui[/bold white] is set, a bec-designer .ui file will also be created. If \
[bold white]open_editor[/bold white] is additionally set, the .ui file will be opened in \
bec-designer and the compiled python version will be updated when changes are made and saved."""
    if (formatted_name := name.lower().replace("-", "_")) != name:
        logger.warning(f"Adjusting widget name from {name} to {formatted_name}")
    if not formatted_name.isidentifier():
        logger.error(
            f"{name} is not a valid name for a widget (even after converting to {formatted_name}) - please enter something in snake_case"
        )
        exit(-1)
    logger.info(f"Adding new widget {formatted_name} to the template...")
    try:
        repo = Path(plugin_repo_path())
        plugin_data = existing_data(repo, [ANSWER_KEYS.VERSION, ANSWER_KEYS.WIDGETS])
        if _widget_exists(plugin_data[ANSWER_KEYS.WIDGETS], formatted_name):
            logger.error(f"Widget {formatted_name} already exists!")
            exit(-1)
        plugin_data[ANSWER_KEYS.WIDGETS].append({"name": formatted_name, "use_ui": use_ui})
        copier.run_update(
            repo,
            data=plugin_data,
            defaults=True,
            unsafe=True,
            overwrite=True,
            vcs_ref=plugin_data[ANSWER_KEYS.VERSION],
        )
        _commit_added_widget(repo, formatted_name)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("exiting...")
        exit(-1)
    logger.success(f"Added widget {formatted_name}!")
    if open_editor:
        open_and_watch_ui_editor(formatted_name)



from __future__ import annotations

from bec_widgets.utils.toolbars.actions import MaterialIconAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle, ToolbarComponents
from bec_widgets.utils.toolbars.connections import BundleConnection


def roi_bundle(components: ToolbarComponents) -> ToolbarBundle:
    """
    Creates a toolbar bundle for ROI and crosshair interaction.

    Args:
        components (ToolbarComponents): The components to be added to the bundle.

    Returns:
        ToolbarBundle: The ROI toolbar bundle.
    """
    components.add_safe(
        "crosshair",
        MaterialIconAction(
            icon_name="point_scan",
            tooltip="Show Crosshair",
            checkable=True,
            parent=components.toolbar,
        ),
    )
    components.add_safe(
        "reset_legend",
        MaterialIconAction(
            icon_name="restart_alt",
            tooltip="Reset the position of legend.",
            checkable=False,
            parent=components.toolbar,
        ),
    )
    bundle = ToolbarBundle("roi", components)
    bundle.add_action("crosshair")
    bundle.add_action("reset_legend")
    return bundle


class RoiConnection(BundleConnection):
    """
    Connection class for the ROI toolbar bundle.
    """

    def __init__(self, components: ToolbarComponents, target_widget=None):
        self.bundle_name = "roi"
        self.components = components
        self.target_widget = target_widget
        if not hasattr(self.target_widget, "toggle_crosshair") or not hasattr(
            self.target_widget, "reset_legend"
        ):
            raise AttributeError(
                "Target widget must implement 'toggle_crosshair' and 'reset_legend'."
            )
        super().__init__()
        self._connected = False

    def connect(self):
        self._connected = True
        # Connect the action to the target widget's method
        self.components.get_action_reference("crosshair")().action.toggled.connect(
            self.target_widget.toggle_crosshair
        )
        self.components.get_action_reference("reset_legend")().action.triggered.connect(
            self.target_widget.reset_legend
        )

    def disconnect(self):
        if not self._connected:
            return
        # Disconnect the action from the target widget's method
        self.components.get_action_reference("crosshair")().action.toggled.disconnect(
            self.target_widget.toggle_crosshair
        )
        self.components.get_action_reference("reset_legend")().action.triggered.disconnect(
            self.target_widget.reset_legend
        )



from __future__ import annotations

import os
import traceback
from typing import TYPE_CHECKING, Literal

import black
import isort
import qtmonaco
from bec_lib.logger import bec_logger
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QApplication, QDialog, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_theme_name
from bec_widgets.utils.error_popups import SafeSlot

if TYPE_CHECKING:
    from bec_widgets.widgets.editors.monaco.scan_control_dialog import ScanControlDialog

logger = bec_logger.logger


class MonacoWidget(BECWidget, QWidget):
    """
    A simple Monaco editor widget
    """

    text_changed = Signal(str)
    save_enabled = Signal(bool)
    PLUGIN = True
    ICON_NAME = "code"
    USER_ACCESS = [
        "set_text",
        "get_text",
        "insert_text",
        "delete_line",
        "open_file",
        "set_language",
        "get_language",
        "set_theme",
        "get_theme",
        "set_readonly",
        "set_cursor",
        "current_cursor",
        "set_minimap_enabled",
        "set_vim_mode_enabled",
        "set_lsp_header",
        "get_lsp_header",
        "attach",
        "detach",
        "screenshot",
    ]

    def __init__(
        self, parent=None, config=None, client=None, gui_id=None, init_lsp: bool = True, **kwargs
    ):
        super().__init__(
            parent=parent, client=client, gui_id=gui_id, config=config, theme_update=True, **kwargs
        )
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.editor = qtmonaco.Monaco(self)
        layout.addWidget(self.editor)
        self.setLayout(layout)
        self.editor.text_changed.connect(self.text_changed.emit)
        self.editor.text_changed.connect(self._check_save_status)
        self.editor.initialized.connect(self.apply_theme)
        self.editor.initialized.connect(self._setup_context_menu)
        self.editor.context_menu_action_triggered.connect(self._handle_context_menu_action)
        self._current_file = None
        self._original_content = ""
        self.metadata = {}
        if init_lsp:
            self.editor.update_workspace_configuration(
                {
                    "pylsp": {
                        "plugins": {
                            "pylsp-bec": {"service_config": self.client._service_config.config}
                        }
                    }
                }
            )

    @property
    def current_file(self):
        """
        Get the current file being edited.
        """
        return self._current_file

    def apply_theme(self, theme: str | None = None) -> None:
        """
        Apply the current theme to the Monaco editor.

        Args:
            theme (str, optional): The theme to apply. If None, the current theme will be used.
        """
        if theme is None:
            theme = get_theme_name()
        editor_theme = "vs" if theme == "light" else "vs-dark"
        self.set_theme(editor_theme)

    def set_text(self, text: str, file_name: str | None = None, reset: bool = False) -> None:
        """
        Set the text in the Monaco editor.

        Args:
            text (str): The text to set in the editor.
            file_name (str): Set the file name
            reset (bool): If True, reset the original content to the new text.
        """
        if reset:
            self._current_file = file_name
            self._original_content = text
        else:
            self._current_file = file_name if file_name else self._current_file

        self.editor.set_text(text, uri=file_name)

    def get_text(self) -> str:
        """
        Get the current text from the Monaco editor.
        """
        return self.editor.get_text()

    def format(self) -> None:
        """
        Format the current text in the Monaco editor.
        """
        if not self.editor:
            return
        try:
            content = self.get_text()
            try:
                formatted_content = black.format_str(content, mode=black.Mode(line_length=100))
            except Exception:  # black.NothingChanged or other formatting exceptions
                formatted_content = content

            config = isort.Config(
                profile="black",
                line_length=100,
                multi_line_output=3,
                include_trailing_comma=False,
                known_first_party=["bec_widgets"],
            )
            formatted_content = isort.code(formatted_content, config=config)
            self.set_text(formatted_content, file_name=self.current_file)
        except Exception:
            content = traceback.format_exc()
            logger.info(content)

    def insert_text(self, text: str, line: int | None = None, column: int | None = None) -> None:
        """
        Insert text at the current cursor position or at a specified line and column.

        Args:
            text (str): The text to insert.
            line (int, optional): The line number (1-based) to insert the text at. Defaults to None.
            column (int, optional): The column number (1-based) to insert the text at. Defaults to None.
        """
        self.editor.insert_text(text, line, column)

    def delete_line(self, line: int | None = None) -> None:
        """
        Delete a line in the Monaco editor.

        Args:
            line (int, optional): The line number (1-based) to delete. If None, the current line will be deleted.
        """
        self.editor.delete_line(line)

    def open_file(self, file_name: str) -> None:
        """
        Open a file in the editor.

        Args:
            file_name (str): The path + file name of the file that needs to be displayed.
        """

        if not os.path.exists(file_name):
            raise FileNotFoundError(f"The specified file does not exist: {file_name}")

        with open(file_name, "r", encoding="utf-8") as file:
            content = file.read()
        self.set_text(content, file_name=file_name, reset=True)

    @property
    def modified(self) -> bool:
        """
        Check if the editor content has been modified.
        """
        return self._original_content != self.get_text()

    @SafeSlot(str)
    def _check_save_status(self, _text: str) -> None:
        self.save_enabled.emit(self.modified)

    def set_cursor(
        self,
        line: int,
        column: int = 1,
        move_to_position: Literal[None, "center", "top", "position"] = None,
    ) -> None:
        """
        Set the cursor position in the Monaco editor.

        Args:
            line (int): Line number (1-based).
            column (int): Column number (1-based), defaults to 1.
            move_to_position (Literal[None, "center", "top", "position"], optional): Position to move the cursor to.
        """
        self.editor.set_cursor(line, column, move_to_position)

    def current_cursor(self) -> dict[str, int]:
        """
        Get the current cursor position in the Monaco editor.

        Returns:
            dict[str, int]: A dictionary with keys 'line' and 'column' representing the cursor position.
        """
        return self.editor.current_cursor

    def set_language(self, language: str) -> None:
        """
        Set the programming language for syntax highlighting in the Monaco editor.

        Args:
            language (str): The programming language to set (e.g., "python", "javascript").
        """
        self.editor.set_language(language)

    def get_language(self) -> str:
        """
        Get the current programming language set in the Monaco editor.
        """
        return self.editor.get_language()

    def set_readonly(self, read_only: bool) -> None:
        """
        Set the Monaco editor to read-only mode.

        Args:
            read_only (bool): If True, the editor will be read-only.
        """
        self.editor.set_readonly(read_only)

    def set_theme(self, theme: str) -> None:
        """
        Set the theme for the Monaco editor.

        Args:
            theme (str): The theme to set (e.g., "vs-dark", "light").
        """
        self.editor.set_theme(theme)

    def get_theme(self) -> str:
        """
        Get the current theme of the Monaco editor.
        """
        return self.editor.get_theme()

    def set_minimap_enabled(self, enabled: bool) -> None:
        """
        Enable or disable the minimap in the Monaco editor.

        Args:
            enabled (bool): If True, the minimap will be enabled; otherwise, it will be disabled.
        """
        self.editor.set_minimap_enabled(enabled)

    def set_highlighted_lines(self, start_line: int, end_line: int) -> None:
        """
        Highlight a range of lines in the Monaco editor.

        Args:
            start_line (int): The starting line number (1-based).
            end_line (int): The ending line number (1-based).
        """
        self.editor.set_highlighted_lines(start_line, end_line)

    def clear_highlighted_lines(self) -> None:
        """
        Clear any highlighted lines in the Monaco editor.
        """
        self.editor.clear_highlighted_lines()

    def set_vim_mode_enabled(self, enabled: bool) -> None:
        """
        Enable or disable Vim mode in the Monaco editor.

        Args:
            enabled (bool): If True, Vim mode will be enabled; otherwise, it will be disabled.
        """
        self.editor.set_vim_mode_enabled(enabled)

    def set_lsp_header(self, header: str) -> None:
        """
        Set the LSP (Language Server Protocol) header for the Monaco editor.
        The header is used to provide context for language servers but is not displayed in the editor.

        Args:
            header (str): The LSP header to set.
        """
        self.editor.set_lsp_header(header)

    def get_lsp_header(self) -> str:
        """
        Get the current LSP header set in the Monaco editor.

        Returns:
            str: The LSP header.
        """
        return self.editor.get_lsp_header()

    def _setup_context_menu(self):
        """Setup custom context menu actions for the Monaco editor."""
        # Add the "Insert Scan" action to the context menu
        self.editor.add_action("insert_scan", "Insert Scan", "python")
        # Add the "Format Code" action to the context menu
        self.editor.add_action("format_code", "Format Code", "python")

    def _handle_context_menu_action(self, action_id: str):
        """Handle context menu action triggers."""
        if action_id == "insert_scan":
            self._show_scan_control_dialog()
        elif action_id == "format_code":
            self._format_code()

    def _show_scan_control_dialog(self):
        """Show the scan control dialog and insert the generated scan code."""
        # Import here to avoid circular imports
        from bec_widgets.widgets.editors.monaco.scan_control_dialog import ScanControlDialog

        dialog = ScanControlDialog(self, client=self.client)
        self._run_dialog_and_insert_code(dialog)

    def _run_dialog_and_insert_code(self, dialog: ScanControlDialog):
        """
        Run the dialog and insert the generated scan code if accepted.
        It is a separate method to allow easier testing.

        Args:
            dialog (ScanControlDialog): The scan control dialog instance.
        """
        if dialog.exec_() == QDialog.DialogCode.Accepted:
            scan_code = dialog.get_scan_code()
            if scan_code:
                # Insert the scan code at the current cursor position
                self.insert_text(scan_code)

    def _format_code(self):
        """Format the current code in the editor."""
        self.format()


if __name__ == "__main__":  # pragma: no cover
    qapp = QApplication([])
    widget = MonacoWidget()
    # set the default size
    widget.resize(800, 600)
    widget.set_language("python")
    widget.set_theme("vs-dark")
    widget.editor.set_minimap_enabled(False)
    widget.set_text("""
import numpy as np
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bec_lib.devicemanager import DeviceContainer
    from bec_lib.scans import Scans
    dev: DeviceContainer
    scans: Scans

#######################################
########## User Script ################
#######################################

# This is a comment
def hello_world():
    print("Hello, world!")
            """)
    widget.set_highlighted_lines(1, 3)
    widget.show()
    qapp.exec_()



import os
import sys
from typing import Any

from PIL import Image, ImageChops
from qtpy.QtGui import QPixmap

import bec_widgets

REFERENCE_DIR = os.path.join(
    os.path.dirname(os.path.dirname(bec_widgets.__file__)), "tests/references"
)
REFERENCE_DIR_FAILURES = os.path.join(
    os.path.dirname(os.path.dirname(bec_widgets.__file__)), "tests/reference_failures"
)


def compare_images(image1_path: str, reference_image_path: str):
    """
    Load two images and compare them pixel by pixel

    Args:
        image1_path(str): The path to the first image
        reference_image_path(str): The path to the reference image

    Raises:
        ValueError: If the images are different
    """
    image1 = Image.open(image1_path)
    image2 = Image.open(reference_image_path)
    if image1.size != image2.size:
        raise ValueError("Image size has changed")
    diff = ImageChops.difference(image1, image2)
    if diff.getbbox():
        # copy image1 to the reference directory to upload as artifact
        os.makedirs(REFERENCE_DIR_FAILURES, exist_ok=True)
        image_name = os.path.join(REFERENCE_DIR_FAILURES, os.path.basename(image1_path))
        image1.save(image_name)
        print(f"Image saved to {image_name}")

        raise ValueError("Images are different")


def snap_and_compare(widget: Any, output_directory: str, suffix: str = ""):
    """
    Save a rendering of a widget and compare it to a reference image

    Args:
        widget(any): The widget to render
        output_directory(str): The directory to save the image to
        suffix(str): A suffix to append to the image name

    Raises:
        ValueError: If the images are different

    Examples:
        snap_and_compare(widget, tmpdir, suffix="started")

    """

    if not isinstance(output_directory, str):
        output_directory = str(output_directory)

    os_suffix = sys.platform

    name = (
        f"{widget.__class__.__name__}_{suffix}_{os_suffix}.png"
        if suffix
        else f"{widget.__class__.__name__}_{os_suffix}.png"
    )

    # Save the widget to a pixmap
    test_image_path = os.path.join(output_directory, name)
    pixmap = QPixmap(widget.size())
    widget.render(pixmap)
    pixmap.save(test_image_path)

    try:
        reference_path = os.path.join(REFERENCE_DIR, f"{widget.__class__.__name__}")
        reference_image_path = os.path.join(reference_path, name)

        if not os.path.exists(reference_image_path):
            raise ValueError(f"Reference image not found: {reference_image_path}")

        compare_images(test_image_path, reference_image_path)

    except ValueError:
        image = Image.open(test_image_path)
        os.makedirs(REFERENCE_DIR_FAILURES, exist_ok=True)
        image_name = os.path.join(REFERENCE_DIR_FAILURES, name)
        image.save(image_name)
        print(f"Image saved to {image_name}")
        raise



from __future__ import annotations

from abc import abstractmethod
from typing import Callable

from bec_lib.logger import bec_logger
from qtpy.QtCore import QObject

logger = bec_logger.logger


class BundleConnection(QObject):
    """
    Base class for toolbar bundle connections.

    Provides infrastructure for bidirectional property-toolbar synchronization:
    - Toolbar actions → Widget properties (via action.triggered connections)
    - Widget properties → Toolbar actions (via property_changed signal)
    """

    bundle_name: str

    def __init__(self, parent=None):
        super().__init__(parent)
        self._property_sync_methods: dict[str, Callable] = {}
        self._property_sync_connected = False

    def register_property_sync(self, prop_name: str, sync_method: Callable):
        """
        Register a method to synchronize toolbar state when a property changes.

        This enables automatic toolbar updates when properties are set programmatically,
        restored from QSettings, or changed via RPC.

        Args:
            prop_name: The property name to watch (e.g., "fft", "log", "x_grid")
            sync_method: Method to call when property changes. Should accept the new value
                        and update toolbar state (typically with signals blocked to prevent loops)

        Example:
            def _sync_fft_toolbar(self, value: bool):
                self.fft_action.blockSignals(True)
                self.fft_action.setChecked(value)
                self.fft_action.blockSignals(False)

            self.register_property_sync("fft", self._sync_fft_toolbar)
        """
        self._property_sync_methods[prop_name] = sync_method

    def _resolve_action(self, action_like):
        if hasattr(action_like, "action"):
            return action_like.action
        return action_like

    def register_checked_action_sync(self, prop_name: str, action_like):
        """
        Register a property sync for a checkable QAction (or wrapper with .action).

        This reduces boilerplate for simple boolean → checked state updates.
        """
        qt_action = self._resolve_action(action_like)

        def _sync_checked(value):
            qt_action.blockSignals(True)
            try:
                qt_action.setChecked(bool(value))
            finally:
                qt_action.blockSignals(False)

        self.register_property_sync(prop_name, _sync_checked)

    def connect_property_sync(self, target_widget):
        """
        Connect to target widget's property_changed signal for automatic toolbar sync.

        Call this in your connect() method after registering all property syncs.

        Args:
            target_widget: The widget to monitor for property changes
        """
        if self._property_sync_connected:
            return

        if hasattr(target_widget, "property_changed"):
            target_widget.property_changed.connect(self._on_property_changed)
            self._property_sync_connected = True
        else:
            logger.warning(
                f"{target_widget.__class__.__name__} does not have property_changed signal. "
                "Property-toolbar sync will not work."
            )

    def disconnect_property_sync(self, target_widget):
        """
        Disconnect from target widget's property_changed signal.

        Call this in your disconnect() method.

        Args:
            target_widget: The widget to stop monitoring
        """
        if not self._property_sync_connected:
            return

        if hasattr(target_widget, "property_changed"):
            try:
                target_widget.property_changed.disconnect(self._on_property_changed)
            except (RuntimeError, TypeError):
                # Signal already disconnected or connection doesn't exist
                pass
            self._property_sync_connected = False

    def _on_property_changed(self, prop_name: str, value):
        """
        Internal handler for property changes.

        Calls the registered sync method for the changed property.
        """
        if prop_name in self._property_sync_methods:
            try:
                self._property_sync_methods[prop_name](value)
            except Exception as e:
                logger.error(
                    f"Error syncing toolbar for property '{prop_name}': {e}", exc_info=True
                )

    @abstractmethod
    def connect(self):
        """
        Connects the bundle to the target widget or application.
        This method should be implemented by subclasses to define how the bundle interacts with the target.

        Subclasses should call connect_property_sync(target_widget) if property sync is needed.
        """

    @abstractmethod
    def disconnect(self):
        """
        Disconnects the bundle from the target widget or application.
        This method should be implemented by subclasses to define how to clean up connections.

        Subclasses should call disconnect_property_sync(target_widget) if property sync was connected.
        """



import uuid
from abc import abstractmethod
from typing import Callable, TypedDict

from bec_lib.device import Positioner
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.messages import ScanQueueMessage
from qtpy.QtWidgets import (
    QDialog,
    QDoubleSpinBox,
    QGroupBox,
    QLabel,
    QLineEdit,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.widgets.control.device_control.position_indicator.position_indicator import (
    PositionIndicator,
)
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import BECDeviceFilter
from bec_widgets.widgets.control.device_input.device_line_edit.device_line_edit import (
    DeviceLineEdit,
)
from bec_widgets.widgets.utility.spinner.spinner import SpinnerWidget

logger = bec_logger.logger


class DeviceUpdateUIComponents(TypedDict):
    spinner: SpinnerWidget
    setpoint: QLineEdit
    readback: QLabel
    position_indicator: PositionIndicator
    step_size: QDoubleSpinBox
    device_box: QGroupBox
    stop: QPushButton
    tweak_increase: QPushButton
    tweak_decrease: QPushButton
    units: QLabel


class PositionerBoxBase(BECWidget, QWidget):
    """Contains some core logic for positioner box widgets"""

    current_path = ""
    ICON_NAME = "switch_right"
    RPC = False

    def __init__(self, parent=None, **kwargs):
        """Initialize the PositionerBox widget.

        Args:
            parent: The parent widget.
            device (Positioner): The device to control.
        """
        super().__init__(parent=parent, **kwargs)
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)
        self._dialog = None
        self.get_bec_shortcuts()

    def _check_device_is_valid(self, device: str):
        """Check if the device is a positioner

        Args:
            device (str): The device name
        """
        if device not in self.dev:
            logger.info(f"Device {device} not found in the device list")
            return False
        if not isinstance(self.dev[device], Positioner):
            logger.info(f"Device {device} is not a positioner")
            return False
        return True

    @abstractmethod
    def _device_ui_components(self, device: str) -> DeviceUpdateUIComponents: ...

    def _init_device(
        self,
        device: str,
        position_emit: Callable[[float], None],
        limit_update: Callable[[tuple[float, float]], None],
    ):
        """Init the device view and readback"""
        if not self._check_device_is_valid(device):
            return

        data = self.dev[device].read(cached=True)
        self._on_device_readback(
            device,
            self._device_ui_components(device),
            {"signals": data},
            {},
            position_emit,
            limit_update,
        )

        ui = self._device_ui_components(device)
        if not ui.get("units"):
            return

        try:
            egu = f"[{self.dev[device].egu()}]"
        except Exception:
            egu = ""

        if egu:
            ui["units"].setVisible(True)
            ui["units"].setText(egu)
        else:
            ui["units"].setVisible(False)

    def _stop_device(self, device: str):
        """Stop call"""
        request_id = str(uuid.uuid4())
        params = {"device": device, "rpc_id": request_id, "func": "stop", "args": [], "kwargs": {}}
        msg = ScanQueueMessage(
            scan_type="device_rpc",
            parameter=params,
            queue="emergency",
            metadata={"RID": request_id, "response": False},
        )
        self.client.connector.send(MessageEndpoints.scan_queue_request(self.client.username), msg)

    # pylint: disable=unused-argument
    def _on_device_readback(
        self,
        device: str,
        ui_components: DeviceUpdateUIComponents,
        msg_content: dict,
        metadata: dict,
        position_emit: Callable[[float], None],
        limit_update: Callable[[tuple[float, float]], None],
    ):
        signals = msg_content.get("signals", {})
        # pylint: disable=protected-access
        hinted_signals = self.dev[device]._hints
        precision = getattr(self.dev[device], "precision", 8)
        try:
            precision = int(precision)
        except (TypeError, ValueError):
            precision = int(8)

        spinner = ui_components["spinner"]
        position_indicator = ui_components["position_indicator"]
        readback = ui_components["readback"]
        setpoint = ui_components["setpoint"]

        readback_val = None
        setpoint_val = None

        if len(hinted_signals) == 1:
            signal = hinted_signals[0]
            readback_val = signals.get(signal, {}).get("value")

        for setpoint_signal in ["setpoint", "user_setpoint"]:
            setpoint_val = signals.get(f"{device}_{setpoint_signal}", {}).get("value")
            if setpoint_val is not None:
                break

        if f"{device}_motor_done_move" in signals:
            is_moving = not signals[f"{device}_motor_done_move"].get("value")
        elif f"{device}_motor_is_moving" in signals:
            is_moving = signals[f"{device}_motor_is_moving"].get("value")
        else:
            is_moving = None

        if is_moving is not None:
            spinner.setVisible(True)
            if is_moving:
                spinner.start()
                spinner.setToolTip("Device is moving")
            else:
                spinner.stop()
                spinner.setToolTip("Device is idle")
        else:
            spinner.setVisible(False)

        if readback_val is not None:
            text = f"{readback_val:.{precision}f}"
            readback.setText(text)
            position_emit(readback_val)

        if setpoint_val is not None:
            text = f"{setpoint_val:.{precision}f}"
            setpoint.setText(text)

        limits = self.dev[device].limits
        limit_update(limits)
        if limits is not None and readback_val is not None and limits[0] != limits[1]:
            pos = (readback_val - limits[0]) / (limits[1] - limits[0])
            position_indicator.set_value(pos)

    @staticmethod
    def _update_limits_ui(limits: tuple[float, float], position_indicator, setpoint_validator):
        if limits is not None and limits[0] != limits[1]:
            position_indicator.setToolTip(f"Min: {limits[0]}, Max: {limits[1]}")
            setpoint_validator.setRange(limits[0], limits[1])
        else:
            position_indicator.setToolTip("No limits set")
            setpoint_validator.setRange(float("-inf"), float("inf"))

    def _update_device_ui(self, device: str, ui: DeviceUpdateUIComponents):
        ui["device_box"].setTitle(device)
        ui["readback"].setToolTip(f"{device} readback")
        ui["setpoint"].setToolTip(f"{device} setpoint")
        ui["step_size"].setToolTip(f"Step size for {device}")
        precision = getattr(self.dev[device], "precision", 8)
        try:
            precision = int(precision)
        except (TypeError, ValueError):
            precision = int(8)
        ui["step_size"].setDecimals(precision)
        ui["step_size"].setValue(10**-precision * 10)

    def _swap_readback_signal_connection(self, slot, old_device, new_device):
        self.bec_dispatcher.disconnect_slot(slot, MessageEndpoints.device_readback(old_device))
        self.bec_dispatcher.connect_slot(slot, MessageEndpoints.device_readback(new_device))

    @staticmethod
    def _toggle_enable_buttons(ui: DeviceUpdateUIComponents, enable: bool) -> None:
        """Toggle enable/disable on available buttons

        Args:
            enable (bool): Enable buttons
        """
        ui["tweak_increase"].setEnabled(enable)
        ui["tweak_decrease"].setEnabled(enable)
        ui["stop"].setEnabled(enable)
        ui["setpoint"].setEnabled(enable)
        ui["step_size"].setEnabled(enable)

    def _on_device_change(
        self,
        old_device: str,
        new_device: str,
        position_emit: Callable[[float], None],
        limit_update: Callable[[tuple[float, float]], None],
        on_device_readback: Callable,
        ui: DeviceUpdateUIComponents,
    ):
        logger.info(f"Device changed from {old_device} to {new_device}")
        self._toggle_enable_buttons(ui, True)
        self._init_device(new_device, position_emit, limit_update)
        self._swap_readback_signal_connection(on_device_readback, old_device, new_device)
        self._update_device_ui(new_device, ui)

    def _open_dialog_selection(self, set_positioner: Callable):
        def _ods():
            """Open dialog window for positioner selection"""
            self._dialog = QDialog(self)
            self._dialog.setWindowTitle("Positioner Selection")
            layout = QVBoxLayout()
            line_edit = DeviceLineEdit(
                self, client=self.client, device_filter=[BECDeviceFilter.POSITIONER]
            )
            line_edit.textChanged.connect(set_positioner)
            layout.addWidget(line_edit)
            close_button = QPushButton("Close")
            close_button.clicked.connect(self._dialog.accept)
            layout.addWidget(close_button)
            self._dialog.setLayout(layout)
            self._dialog.exec()
            self._dialog = None

        return _ods



from __future__ import annotations

import inspect
import typing
from abc import abstractmethod
from decimal import Decimal
from types import GenericAlias, NoneType, UnionType
from typing import (
    Any,
    Callable,
    Final,
    Generic,
    Iterable,
    Literal,
    NamedTuple,
    Optional,
    OrderedDict,
    Protocol,
    TypeVar,
    get_args,
    runtime_checkable,
)

from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from qtpy import QtCore
from qtpy.QtCore import QSize, Qt, Signal  # type: ignore
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import (
    QApplication,
    QButtonGroup,
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QRadioButton,
    QSizePolicy,
    QSpinBox,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.widget_io import WidgetIO
from bec_widgets.widgets.editors.dict_backed_table import DictBackedTable
from bec_widgets.widgets.editors.scan_metadata._util import (
    clearable_required,
    field_default,
    field_limits,
    field_maxlen,
    field_minlen,
    field_precision,
)
from bec_widgets.widgets.utility.toggle.toggle import ToggleSwitch

logger = bec_logger.logger


class FormItemSpec(BaseModel):
    """
    The specification for an item in a dynamically generated form. Uses a pydantic FieldInfo
    to store most annotation info, since one of the main purposes is to store data for
    forms generated from pydantic models, but can also be composed from other sources or by hand.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    item_type: type | UnionType | GenericAlias | Optional[Any]
    name: str
    info: FieldInfo = FieldInfo()
    pretty_display: bool = Field(
        default=False,
        description="Whether to use a pretty display for the widget. Defaults to False. If True, disables the widget, doesn't add a clear button, and adapts the stylesheet for non-editable display.",
    )

    @field_validator("item_type", mode="before")
    @classmethod
    def _validate_type(cls, v):
        allowed_primitives = [str, int, float, bool]
        if isinstance(v, (type, UnionType)):
            return v
        if isinstance(v, GenericAlias):
            if v.__origin__ in [list, dict, set] and all(
                arg in allowed_primitives for arg in v.__args__
            ):
                return v
            raise ValueError(
                f"Generics of type {v} are not supported - only lists, dicts and sets of primitive types {allowed_primitives}"
            )
        if type(v) is type(Literal[""]):  # _LiteralGenericAlias is not exported from typing
            arg_types = set(type(arg) for arg in v.__args__)
            if len(arg_types) != 1:
                raise ValueError("Mixtures of literal types are not supported!")
            if (t := arg_types.pop()) in allowed_primitives:
                return t
            raise ValueError(f"Literals of type {t} are not supported")


class ClearableBoolEntry(QWidget):
    stateChanged = Signal()

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._layout = QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)
        self._layout.setSizeConstraint(QLayout.SizeConstraint.SetFixedSize)
        self._entry = QButtonGroup()
        self._true = QRadioButton("true", parent=self)
        self._false = QRadioButton("false", parent=self)
        for button in [self._true, self._false]:
            self._layout.addWidget(button)
            self._entry.addButton(button)
            button.toggled.connect(self.stateChanged)

    def clear(self):
        self._entry.setExclusive(False)
        self._true.setChecked(False)
        self._false.setChecked(False)
        self._entry.setExclusive(True)

    def isChecked(self) -> bool | None:
        if not self._true.isChecked() and not self._false.isChecked():
            return None
        return self._true.isChecked()

    def setChecked(self, value: bool | None):
        if value is None:
            self.clear()
        elif value:
            self._true.setChecked(True)
            self._false.setChecked(False)
        else:
            self._true.setChecked(False)
            self._false.setChecked(True)

    def setToolTip(self, tooltip: str):
        self._true.setToolTip(tooltip)
        self._false.setToolTip(tooltip)


DynamicFormItemType = str | int | float | Decimal | bool | dict | list | None


class DynamicFormItem(QWidget):
    valueChanged = Signal()

    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        """
        Initializes the form item widget.

        Args:
            parent (QWidget | None, optional): The parent widget. Defaults to None.
            spec (FormItemSpec): The specification for the form item.
        """
        super().__init__(parent)
        self._spec = spec
        self._layout = QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSizeConstraint(QLayout.SizeConstraint.SetMaximumSize)
        self._default = field_default(self._spec.info)
        self._desc = self._spec.info.description
        self.setLayout(self._layout)
        self._add_main_widget()
        # Sadly, QWidget and ABC are not compatible
        assert isinstance(self._main_widget, QWidget), "Please set a widget in _add_main_widget()"  # type: ignore
        self._main_widget.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        if not spec.pretty_display:
            if clearable_required(spec.info):
                self._add_clear_button()
        else:
            self._set_pretty_display()

    @abstractmethod
    def getValue(self) -> DynamicFormItemType: ...

    @abstractmethod
    def setValue(self, value): ...

    @abstractmethod
    def _add_main_widget(self) -> None:
        self._main_widget: QWidget
        """Add the main data entry widget to self._main_widget and apply any
        constraints from the field info"""

    @SafeSlot()
    def clear(self, *_):
        return

    def _set_pretty_display(self):
        self.setEnabled(False)
        if button := getattr(self, "_clear_button", None):
            button.setVisible(False)

    def _describe(self, pad=" "):
        return pad + (self._desc if self._desc else "")

    def _add_clear_button(self):
        self._clear_button = QToolButton()
        self._clear_button.setIcon(
            material_icon(icon_name="close", size=(10, 10), convert_to_pixmap=False)
        )
        self._layout.addWidget(self._clear_button)
        # the widget added in _add_main_widget must implement .clear() if value is not required
        self._clear_button.setToolTip("Clear value or reset to default.")
        self._clear_button.clicked.connect(self.clear)  # type: ignore

    def _value_changed(self, *_, **__):
        self.valueChanged.emit()

    def teardown(self):
        self._layout.deleteLater()
        self._layout.removeWidget(self._main_widget)
        self._main_widget.deleteLater()
        self._main_widget = None


class StrFormItem(DynamicFormItem):
    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.textChanged.connect(self._value_changed)
        if spec.info.description:
            self._main_widget.setPlaceholderText(spec.info.description)

    def _add_main_widget(self) -> None:
        self._main_widget = QLineEdit()
        self._layout.addWidget(self._main_widget)
        min_length, max_length = (field_minlen(self._spec.info), field_maxlen(self._spec.info))
        if max_length:
            self._main_widget.setMaxLength(max_length)
        self._main_widget.setToolTip(
            f"(length min: {min_length} max: {max_length}){self._describe()}"
        )
        if self._default:
            self._main_widget.setText(self._default)
            self._add_clear_button()

    def getValue(self):
        if self._main_widget.text() == "":
            return self._default
        return self._main_widget.text()

    def setValue(self, value: str):
        if value is None:
            return self._main_widget.setText("")
        self._main_widget.setText(str(value))


class IntFormItem(DynamicFormItem):
    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.textChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        self._main_widget = QSpinBox()
        self._layout.addWidget(self._main_widget)
        min_, max_ = field_limits(self._spec.info, int)
        self._main_widget.setMinimum(min_)
        self._main_widget.setMaximum(max_)
        self._main_widget.setToolTip(f"(range {min_} to {max_}){self._describe()}")
        if self._default is not None:
            self._main_widget.setValue(self._default)
            self._add_clear_button()
        else:
            self._main_widget.clear()

    def getValue(self):
        if self._main_widget.text() == "":
            return self._default
        return self._main_widget.value()

    def setValue(self, value: int):
        if value is None:
            self._main_widget.clear()
        self._main_widget.setValue(value)


class FloatDecimalFormItem(DynamicFormItem):
    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.textChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        precision = field_precision(self._spec.info)
        self._main_widget = QDoubleSpinBox()
        self._layout.addWidget(self._main_widget)
        min_, max_ = field_limits(self._spec.info, float, precision)
        self._main_widget.setMinimum(min_)
        self._main_widget.setMaximum(max_)
        if precision:
            self._main_widget.setDecimals(precision)
        minstr = f"{float(min_):.3f}" if abs(min_) <= 1000 else f"{float(min_):.3e}"
        maxstr = f"{float(max_):.3f}" if abs(max_) <= 1000 else f"{float(max_):.3e}"
        self._main_widget.setToolTip(f"(range {minstr} to {maxstr}){self._describe()}")
        if self._default is not None:
            self._main_widget.setValue(self._default)
            self._add_clear_button()
        else:
            self._main_widget.clear()

    def getValue(self):
        if self._main_widget.text() == "":
            return self._default
        return self._main_widget.value()

    def setValue(self, value: float | Decimal):
        if value is None:
            self._main_widget.clear()
        self._main_widget.setValue(float(value))


class BoolFormItem(DynamicFormItem):
    def __init__(self, *, parent: QWidget | None = None, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.stateChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        if clearable_required(self._spec.info):
            self._main_widget = ClearableBoolEntry()
        else:
            self._main_widget = QCheckBox()
        self._layout.addWidget(self._main_widget)
        self._main_widget.setToolTip(self._describe(""))
        self._main_widget.setChecked(self._default)  # type: ignore # if there is no default then it will be ClearableBoolEntry and can be set with None

    def getValue(self):
        return self._main_widget.isChecked()

    def setValue(self, value):
        self._main_widget.setChecked(value)


class BoolToggleFormItem(BoolFormItem):
    def __init__(self, *, parent: QWidget | None = None, spec: FormItemSpec) -> None:
        if spec.info.default is PydanticUndefined:
            spec.info.default = False
        super().__init__(parent=parent, spec=spec)

    def _add_main_widget(self) -> None:
        self._main_widget = ToggleSwitch()
        self._layout.addWidget(self._main_widget)
        self._main_widget.setToolTip(self._describe(""))
        if self._default is not None:
            self._main_widget.setChecked(self._default)


class DictFormItem(DynamicFormItem):
    def __init__(self, *, parent: QWidget | None = None, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.data_changed.connect(self._value_changed)
        if spec.info.default is not PydanticUndefined:
            self._main_widget.set_default(spec.info.default)

    def _set_pretty_display(self):
        self._main_widget.set_button_visibility(False)
        super()._set_pretty_display()

    def _add_main_widget(self) -> None:
        self._main_widget = DictBackedTable(self, [])
        self._layout.addWidget(self._main_widget)
        self._main_widget.setToolTip(self._describe(""))

    def getValue(self):
        return self._main_widget.dump_dict()

    def setValue(self, value):
        self._main_widget.replace_data(value)


_IW = TypeVar("_IW", bound=int | float | str)


class _ItemAndWidgetType(NamedTuple, Generic[_IW]):
    item: type[_IW]
    widget: type[QWidget]
    default: _IW


class ListFormItem(DynamicFormItem):
    def __init__(self, *, parent: QWidget | None = None, spec: FormItemSpec) -> None:
        if spec.info.annotation is list:
            self._types = _ItemAndWidgetType(str, QLineEdit, "")
        elif isinstance(spec.info.annotation, GenericAlias):
            args = set(typing.get_args(spec.info.annotation))
            if args == {str}:
                self._types = _ItemAndWidgetType(str, QLineEdit, "")
            if args == {int}:
                self._types = _ItemAndWidgetType(int, QSpinBox, 0)
            if args == {float} or args == {int, float}:
                self._types = _ItemAndWidgetType(float, QDoubleSpinBox, 0.0)
        else:
            self._types = _ItemAndWidgetType(str, QLineEdit, "")
        super().__init__(parent=parent, spec=spec)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self._main_widget: QListWidget
        self._data = []
        self._min_lines = 2 if spec.pretty_display else 4
        self._repop(self._data)

    def sizeHint(self):
        default = super().sizeHint()
        return QSize(default.width(), QFontMetrics(self.font()).height() * 4)

    def _add_main_widget(self) -> None:
        self._main_widget = QListWidget()
        self._layout.addWidget(self._main_widget)
        self._layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._add_buttons()

    def _add_buttons(self):
        self._button_holder = QWidget()
        self._button_holder.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self._buttons = QVBoxLayout()
        self._buttons.setContentsMargins(0, 0, 0, 0)
        self._button_holder.setLayout(self._buttons)
        self._layout.addWidget(self._button_holder)

        self._add_remove_button_holder = QWidget()
        self._add_remove_button_layout = QHBoxLayout()
        self._add_remove_button_layout.setContentsMargins(0, 0, 0, 0)
        self._add_remove_button_holder.setLayout(self._add_remove_button_layout)

        self._add_button = QPushButton("+")
        self._add_button.setMinimumHeight(15)
        self._add_button.setToolTip("add a new row")
        self._remove_button = QPushButton("-")
        self._remove_button.setMinimumHeight(15)
        self._remove_button.setToolTip("delete the focused row (if any)")
        self._add_button.clicked.connect(self._add_row)
        self._remove_button.clicked.connect(self._delete_row)

        self._buttons.addWidget(self._add_remove_button_holder)
        self._add_remove_button_layout.addWidget(self._add_button)
        self._add_remove_button_layout.addWidget(self._remove_button)

    def _set_pretty_display(self):
        super()._set_pretty_display()
        self._button_holder.setHidden(True)

    def _repop(self, data):
        self._main_widget.clear()
        for val in data:
            self._add_list_item(val)
        self.scale_to_data()

    def _add_data_item(self, val=None):
        val = val or self._types.default
        self._data.append(val)
        self._add_list_item(val)
        self._repop(self._data)

    def _item_height(self):
        return int(QFontMetrics(self.font()).height() * 1.5)

    def _add_list_item(self, val):
        item = QListWidgetItem(self._main_widget)
        item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEditable)
        item_widget = self._types.widget(parent=self)
        item_widget.setMinimumHeight(self._item_height())
        self._main_widget.setGridSize(QSize(0, self._item_height()))
        if (layout := item_widget.layout()) is not None:
            layout.setContentsMargins(0, 0, 0, 0)
        WidgetIO.set_value(item_widget, val)
        self._main_widget.setItemWidget(item, item_widget)
        self._main_widget.addItem(item)
        WidgetIO.connect_widget_change_signal(item_widget, self._update)
        return item_widget

    def _update(self, _, value, *args):
        self._data[self._main_widget.currentRow()] = value

    @SafeSlot()
    def _add_row(self):
        self._add_data_item(self._types.default)
        self._repop(self._data)

    @SafeSlot()
    def _delete_row(self):
        if selected := self._main_widget.currentItem():
            self._main_widget.removeItemWidget(selected)
            row = self._main_widget.currentRow()
            self._main_widget.takeItem(row)
            self._data.pop(row)
        self._repop(self._data)

    @SafeSlot()
    def clear(self):
        self._repop([])

    def getValue(self):
        return self._data

    def setValue(self, value: Iterable):
        if set(map(type, value)) | {self._types.item} != {self._types.item}:
            raise ValueError(f"This widget only accepts items of type {self._types.item}")
        self._data = list(value)
        self._repop(self._data)

    def set_max_height_in_lines(self, lines: int):
        outer_inc = 1 if self._spec.pretty_display else 3
        self._main_widget.setFixedHeight(self._item_height() * max(lines, self._min_lines))
        self._button_holder.setFixedHeight(self._item_height() * (max(lines, self._min_lines) + 1))
        self.setFixedHeight(self._item_height() * (max(lines, self._min_lines) + outer_inc))

    def scale_to_data(self, *_):
        self.set_max_height_in_lines(self._main_widget.count() + 1)


class SetFormItem(ListFormItem):
    def _add_main_widget(self) -> None:
        super()._add_main_widget()
        self._add_item_field = self._types.widget()
        self._buttons.addWidget(QLabel("Add new:"))
        self._buttons.addWidget(self._add_item_field)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Minimum)

    @SafeSlot()
    def _add_row(self):
        self._add_data_item(WidgetIO.get_value(self._add_item_field))
        self._repop(self._data)

    def _update(self, _, value, *args):
        if value in self._data:
            return
        return super()._update(_, value, *args)

    def _add_data_item(self, val=None):
        val = val or self._types.default
        if val == self._types.default or val in self._data:
            return
        self._data.append(val)
        self._add_list_item(val)

    def _add_list_item(self, val):
        item_widget = super()._add_list_item(val)
        if isinstance(item_widget, QLineEdit):
            item_widget.setReadOnly(True)
        return item_widget

    def getValue(self):
        return set(self._data)

    def setValue(self, value: set):
        return super().setValue(set(value))


class StrLiteralFormItem(DynamicFormItem):
    def _add_main_widget(self) -> None:
        self._main_widget = QComboBox()
        self._options = get_args(self._spec.info.annotation)
        for opt in self._options:
            self._main_widget.addItem(opt)
        self._layout.addWidget(self._main_widget)

    def getValue(self):
        if self._main_widget.currentIndex() == -1:
            return None
        return self._main_widget.currentText()

    def setValue(self, value: str | None):
        if value is None:
            self.clear()
            return
        for i in range(self._main_widget.count()):
            if self._main_widget.itemText(i) == value:
                self._main_widget.setCurrentIndex(i)
                return
        raise ValueError(f"Cannot set value: {value}, options are: {self._options}")

    def clear(self):
        self._main_widget.setCurrentIndex(-1)


class OptionalStrLiteralFormItem(StrLiteralFormItem):
    def _add_main_widget(self) -> None:
        self._main_widget = QComboBox()
        self._options = get_args(get_args(self._spec.info.annotation)[0])
        for opt in self._options:
            self._main_widget.addItem(opt)
        self._layout.addWidget(self._main_widget)


WidgetTypeRegistry = OrderedDict[str, tuple[Callable[[FormItemSpec], bool], type[DynamicFormItem]]]


@runtime_checkable
class _ItemTypeFn(Protocol):
    def __call__(self, spec: FormItemSpec) -> type[DynamicFormItem]: ...


WidgetTypeRegistry = OrderedDict[
    str, tuple[Callable[[FormItemSpec], bool], type[DynamicFormItem] | _ItemTypeFn]
]


def _is_string_literal(t: type):
    return type(t) is type(Literal[""]) and set(type(arg) for arg in get_args(t)) == {str}


def _is_optional_string_literal(t: type):
    if not hasattr(t, "__args__"):
        return False
    if len(t.__args__) != 2:
        return False
    if _is_string_literal(t.__args__[0]) and t.__args__[1] is NoneType:
        return True
    return False


DEFAULT_WIDGET_TYPES: Final[WidgetTypeRegistry] = OrderedDict() | {
    # dict literals are ordered already but TypedForm subclasses may modify coppies of this dict
    # and delete/insert keys or change the order
    "literal_str": (lambda spec: _is_string_literal(spec.info.annotation), StrLiteralFormItem),
    "optional_literal_str": (
        lambda spec: _is_optional_string_literal(spec.info.annotation),
        OptionalStrLiteralFormItem,
    ),
    "str": (lambda spec: spec.item_type in [str, str | None, None], StrFormItem),
    "int": (lambda spec: spec.item_type in [int, int | None], IntFormItem),
    "float_decimal": (
        lambda spec: spec.item_type in [float, float | None, Decimal, Decimal | None],
        FloatDecimalFormItem,
    ),
    "bool": (lambda spec: spec.item_type in [bool, bool | None], BoolFormItem),
    "dict": (
        lambda spec: spec.item_type in [dict, dict | None]
        or (isinstance(spec.item_type, GenericAlias) and spec.item_type.__origin__ is dict),
        DictFormItem,
    ),
    "list": (
        lambda spec: spec.item_type in [list, list | None]
        or (isinstance(spec.item_type, GenericAlias) and spec.item_type.__origin__ is list),
        ListFormItem,
    ),
    "set": (
        lambda spec: spec.item_type in [set, set | None]
        or (isinstance(spec.item_type, GenericAlias) and spec.item_type.__origin__ is set),
        SetFormItem,
    ),
}


def widget_from_type(
    spec: FormItemSpec, widget_types: WidgetTypeRegistry | None = None
) -> type[DynamicFormItem]:
    widget_types = widget_types or DEFAULT_WIDGET_TYPES
    for predicate, widget_type in widget_types.values():
        if predicate(spec):
            if inspect.isclass(widget_type) and issubclass(widget_type, DynamicFormItem):
                return widget_type
            return widget_type(spec)

    logger.warning(
        f"Type {spec.item_type=} / {spec.info.annotation=} is not (yet) supported in dynamic form creation."
    )
    return StrFormItem


if __name__ == "__main__":  # pragma: no cover

    class TestModel(BaseModel):
        value0: set = Field(set(["a", "b"]))
        value1: str | None = Field(None)
        value2: bool | None = Field(None)
        value3: bool = Field(True)
        value4: int = Field(123)
        value5: int | None = Field()
        value6: list[int] = Field()
        value7: list = Field()
        literal: Literal["a", "b", "c"]
        nullable_literal: Literal["a", "b", "c"] | None = None

    app = QApplication([])
    w = QWidget()
    layout = QGridLayout()
    w.setLayout(layout)
    items = []
    for i, (field_name, info) in enumerate(TestModel.model_fields.items()):
        spec = FormItemSpec(item_type=info.annotation, name=field_name, info=info)
        layout.addWidget(QLabel(field_name), i, 0)
        widg = widget_from_type(spec)(spec=spec)
        items.append(widg)
        layout.addWidget(widg, i, 1)

    items[6].setValue([1, 2, 3, 4])
    items[7].setValue(["1", "2", "asdfg", "qwerty"])

    w.show()
    app.exec()



"""Utility functions for experiment selection."""

from datetime import datetime
from typing import Literal

from bec_lib.messages import ExperimentInfoMessage


def format_name(info: dict | ExperimentInfoMessage) -> str:
    """Format the name from the experiment info."""
    info = ExperimentInfoMessage.model_validate(info) if isinstance(info, dict) else info
    firstname = info.firstname
    lastname = info.lastname
    return " ".join(part for part in [firstname, lastname] if part)


def format_schedule(
    schedule: list[dict[Literal["start", "end"], str]] | None, as_datetime: bool = False
) -> tuple[str, str] | tuple[datetime | None, datetime | None]:
    """Format the schedule information to display start and end times."""
    if not schedule:
        return (None, None) if as_datetime else ("", "")
    start, end = _pick_schedule_entry(schedule)
    if as_datetime:
        return start, end
    return format_datetime(start), format_datetime(end)


def _pick_schedule_entry(
    schedule: list[dict[Literal["start", "end"], str]],
) -> tuple[datetime | None, datetime | None]:
    """Pick the most relevant schedule entry based on the current time."""
    now = datetime.now()
    candidates = []
    for item in schedule:
        if not item:
            continue
        start_raw = item.get("start")
        parsed = _parse_schedule_start(start_raw)
        if parsed is None:
            continue
        candidates.append((parsed, item))

    if not candidates:
        return None, None

    future = [entry for entry in candidates if entry[0] >= now]
    pool = future or candidates
    chosen_start, chosen_item = min(pool, key=lambda entry: abs(entry[0] - now))
    end_raw = chosen_item.get("end")
    return chosen_start, _parse_schedule_start(end_raw)


def _parse_schedule_start(value) -> datetime | None:
    """Parse a schedule start string into a datetime object."""
    if not value:
        return None
    try:
        return datetime.strptime(value, "%d/%m/%Y %H:%M:%S")
    except ValueError:
        return None


def format_datetime(value) -> str:
    if not value:
        return ""
    return value.strftime("%Y-%m-%d %H:%M")



import json
from typing import Literal

import pyqtgraph as pg
from bec_lib.logger import bec_logger
from qtpy.QtCore import QPointF, QSize, Qt
from qtpy.QtWidgets import QHBoxLayout, QLabel, QVBoxLayout, QWidget

from bec_widgets.utils import Colors
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty
from bec_widgets.utils.settings_dialog import SettingsDialog
from bec_widgets.utils.toolbars.actions import MaterialIconAction
from bec_widgets.utils.toolbars.toolbar import ModularToolBar
from bec_widgets.widgets.containers.main_window.addons.hover_widget import WidgetTooltip
from bec_widgets.widgets.progress.ring_progress_bar.ring import Ring
from bec_widgets.widgets.progress.ring_progress_bar.ring_progress_settings_cards import RingSettings

logger = bec_logger.logger


class RingProgressContainerWidget(QWidget):
    """
    A container widget for the Ring Progress Bar widget.
    It holds the rings and manages their layout and painting.
    """

    def __init__(self, parent: QWidget | None = None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.rings: list[Ring] = []
        self.gap = 20  # Gap between rings
        self.color_map: str = "turbo"
        self._hovered_ring: Ring | None = None
        self._last_hover_global_pos = None
        self._hover_tooltip_label = QLabel()
        self._hover_tooltip_label.setWordWrap(True)
        self._hover_tooltip_label.setTextFormat(Qt.TextFormat.PlainText)
        self._hover_tooltip_label.setMaximumWidth(260)
        self._hover_tooltip_label.setStyleSheet("font-size: 12px;")
        self._hover_tooltip = WidgetTooltip(self._hover_tooltip_label)
        self.setLayout(QHBoxLayout())
        self.setMouseTracking(True)
        self.initialize_bars()
        self.initialize_center_label()

    @property
    def num_bars(self) -> int:
        return len(self.rings)

    def initialize_bars(self):
        """
        Initialize the progress bars.
        """
        for _ in range(self.num_bars):
            self.add_ring()

        if self.color_map:
            self.set_colors_from_map(self.color_map)

    def add_ring(self, config: dict | None = None) -> Ring:
        """
        Add a new ring to the container.

        Args:
            config(dict | None): Optional configuration dictionary for the ring.

        Returns:
            Ring: The newly added ring object.
        """
        ring = Ring(parent=self)
        ring.setGeometry(self.rect())
        ring.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, True)
        ring.gap = self.gap * len(self.rings)
        ring.set_value(0)
        self.rings.append(ring)
        if config:
            # We have to first get the link_colors property before loading the settings
            # While this is an ugly hack, we do not have control over the order of properties
            # being set when loading.
            ring.link_colors = config.pop("link_colors", True)
            ring.load_settings(config)
        if self.color_map:
            self.set_colors_from_map(self.color_map)
        ring.show()
        ring.raise_()
        self.update()
        return ring

    def remove_ring(self, index: int | None = None):
        """
        Remove a ring from the container.

        Args:
            index(int | None): Index of the ring to remove. If None, removes the last ring.
        """
        if self.num_bars == 0:
            return
        if index is None:
            index = self.num_bars - 1
        index = self._validate_index(index)
        ring = self.rings[index]
        if ring is self._hovered_ring:
            self._hovered_ring = None
            self._last_hover_global_pos = None
            self._hover_tooltip.hide()
        ring.cleanup()
        ring.close()
        ring.deleteLater()
        self.rings.pop(index)
        # Update gaps for remaining rings
        for i, r in enumerate(self.rings):
            r.gap = self.gap * i
        self.update()

    def initialize_center_label(self):
        """
        Initialize the center label.
        """
        layout = self.layout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.center_label = QLabel("", parent=self)
        self.center_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.center_label.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents, True)
        layout.addWidget(self.center_label)

    def _calculate_minimum_size(self):
        """
        Calculate the minimum size of the widget.
        """
        if not self.rings:
            return QSize(10, 10)
        ring_widths = self.get_ring_line_widths()
        total_width = sum(ring_widths) + self.gap * (self.num_bars - 1)
        diameter = max(total_width * 2, 50)

        return QSize(diameter, diameter)

    def get_ring_line_widths(self):
        """
        Get the line widths of the rings.
        """
        if not self.rings:
            return [10]
        ring_widths = [ring.config.line_width for ring in self.rings]
        return ring_widths

    def get_max_ring_size(self) -> int:
        """
        Get the size of the rings.
        """
        if not self.rings:
            return 10
        ring_widths = self.get_ring_line_widths()
        return max(ring_widths)

    def sizeHint(self):
        min_size = self._calculate_minimum_size()
        return min_size

    def resizeEvent(self, event):
        """
        Handle resize events to update ring geometries.
        """
        super().resizeEvent(event)
        for ring in self.rings:
            ring.setGeometry(self.rect())

    def enterEvent(self, event):
        self.setMouseTracking(True)
        super().enterEvent(event)

    def mouseMoveEvent(self, event):
        pos = event.position() if hasattr(event, "position") else QPointF(event.pos())
        self._last_hover_global_pos = (
            event.globalPosition().toPoint()
            if hasattr(event, "globalPosition")
            else event.globalPos()
        )
        ring = self._ring_at_pos(pos)
        self._set_hovered_ring(ring, event)
        super().mouseMoveEvent(event)

    def leaveEvent(self, event):
        self._last_hover_global_pos = None
        self._set_hovered_ring(None, event)
        super().leaveEvent(event)

    def _set_hovered_ring(self, ring: Ring | None, event=None):
        if ring is self._hovered_ring:
            if ring is not None:
                self.refresh_hover_tooltip(ring, event)
            return
        if self._hovered_ring is not None:
            self._hovered_ring.set_hovered(False)
        self._hovered_ring = ring
        if self._hovered_ring is not None:
            self._hovered_ring.set_hovered(True)
            self.refresh_hover_tooltip(self._hovered_ring, event)
        else:
            self._hover_tooltip.hide()

    def _ring_at_pos(self, pos: QPointF) -> Ring | None:
        if not self.rings:
            return None
        size = min(self.width(), self.height())
        if size <= 0:
            return None
        x_offset = (self.width() - size) / 2
        y_offset = (self.height() - size) / 2
        center_x = x_offset + size / 2
        center_y = y_offset + size / 2
        dx = pos.x() - center_x
        dy = pos.y() - center_y
        distance = (dx * dx + dy * dy) ** 0.5

        max_ring_size = self.get_max_ring_size()
        base_radius = (size - 2 * max_ring_size) / 2
        if base_radius <= 0:
            return None

        best_ring: Ring | None = None
        best_delta: float | None = None
        for ring in self.rings:
            radius = base_radius - ring.gap
            if radius <= 0:
                continue
            half_width = ring.config.line_width / 2
            inner = radius - half_width
            outer = radius + half_width
            if inner <= distance <= outer:
                delta = abs(distance - radius)
                if best_delta is None or delta < best_delta:
                    best_delta = delta
                    best_ring = ring

        return best_ring

    def is_ring_hovered(self, ring: Ring) -> bool:
        return ring is self._hovered_ring

    def refresh_hover_tooltip(self, ring: Ring, event=None):
        text = self._build_tooltip_text(ring)
        if event is not None:
            self._last_hover_global_pos = (
                event.globalPosition().toPoint()
                if hasattr(event, "globalPosition")
                else event.globalPos()
            )
        if self._last_hover_global_pos is None:
            return
        self._hover_tooltip_label.setText(text)
        self._hover_tooltip.apply_theme()
        self._hover_tooltip.show_near(self._last_hover_global_pos)

    @staticmethod
    def _build_tooltip_text(ring: Ring) -> str:
        mode = ring.config.mode
        mode_label = {"manual": "Manual", "scan": "Scan progress", "device": "Device"}.get(
            mode, mode
        )

        precision = int(ring.config.precision)
        value = ring.config.value
        min_value = ring.config.min_value
        max_value = ring.config.max_value
        range_span = max(max_value - min_value, 1e-9)
        progress = max(0.0, min(100.0, ((value - min_value) / range_span) * 100))

        lines = [
            f"Mode: {mode_label}",
            f"Progress: {value:.{precision}f} / {max_value:.{precision}f} ({progress:.1f}%)",
        ]
        if min_value != 0:
            lines.append(f"Range: {min_value:.{precision}f} -> {max_value:.{precision}f}")
        if mode == "device" and ring.config.device:
            if ring.config.signal:
                lines.append(f"Device: {ring.config.device}:{ring.config.signal}")
            else:
                lines.append(f"Device: {ring.config.device}")

        return "\n".join(lines)

    def closeEvent(self, event):
        # Ensure the hover tooltip is properly cleaned up when this widget closes
        tooltip = getattr(self, "_hover_tooltip", None)
        if tooltip is not None:
            tooltip.close()
            tooltip.deleteLater()
            self._hover_tooltip = None
        super().closeEvent(event)

    def set_colors_from_map(self, colormap, color_format: Literal["RGB", "HEX"] = "RGB"):
        """
        Set the colors for the progress bars from a colormap.

        Args:
            colormap(str): Name of the colormap.
            color_format(Literal["RGB","HEX"]): Format of the returned colors ('RGB', 'HEX').
        """
        if colormap not in pg.colormap.listMaps():
            raise ValueError(
                f"Colormap '{colormap}' not found in the current installation of pyqtgraph"
            )
        colors = Colors.golden_angle_color(colormap, self.num_bars, color_format)
        self.set_colors_directly(colors)
        self.color_map = colormap
        self.update()

    def set_colors_directly(
        self, colors: list[str | tuple] | str | tuple, bar_index: int | None = None
    ):
        """
        Set the colors for the progress bars directly.

        Args:
            colors(list[str | tuple] | str | tuple): Color(s) for the progress bars. If multiple progress bars are displayed, provide a list of colors for each progress bar.
            bar_index(int): Index of the progress bar to set the color for. If provided, only a single color can be set.
        """
        if bar_index is not None and isinstance(colors, (str, tuple)):
            bar_index = self._validate_index(bar_index)
            self.rings[bar_index].set_color(colors)
        else:
            if isinstance(colors, (str, tuple)):
                colors = [colors]
            colors = self._adjust_list_to_bars(colors)
            for ring, color in zip(self.rings, colors):
                ring.set_color(color)
        self.update()

    def _adjust_list_to_bars(self, items: list) -> list:
        """
        Utility method to adjust the list of parameters to match the number of progress bars.

        Args:
            items(list): List of parameters for the progress bars.

        Returns:
            list: List of parameters for the progress bars.
        """
        if items is None:
            raise ValueError(
                "Items cannot be None. Please provide a list for parameters for the progress bars."
            )
        if not isinstance(items, list):
            items = [items]
        if len(items) < self.num_bars:
            last_item = items[-1]
            items.extend([last_item] * (self.num_bars - len(items)))
        elif len(items) > self.num_bars:
            items = items[: self.num_bars]
        return items

    def _validate_index(self, index: int) -> int:
        """
        Check if the provided index is valid for the number of bars.

        Args:
            index(int): Index to check.
        Returns:
            int: Validated index.
        """
        try:
            self.rings[index]
        except IndexError:
            raise IndexError(f"Index {index} is out of range for {self.num_bars} rings.")
        return index

    def clear_all(self):
        """
        Clear all rings from the widget.
        """
        self._hovered_ring = None
        self._last_hover_global_pos = None
        self._hover_tooltip.hide()
        for ring in self.rings:
            ring.close()
            ring.deleteLater()
        self.rings = []
        self.update()


class RingProgressBar(BECWidget, QWidget):
    ICON_NAME = "track_changes"
    PLUGIN = True
    RPC = True

    USER_ACCESS = [
        *BECWidget.USER_ACCESS,
        "screenshot",
        "rings",
        "add_ring",
        "remove_ring",
        "set_gap",
        "set_center_label",
    ]

    def __init__(self, parent: QWidget | None = None, client=None, **kwargs):
        super().__init__(parent=parent, client=client, theme_update=True, **kwargs)

        self.setWindowTitle("Ring Progress Bar")

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(self.layout)

        self.toolbar = ModularToolBar(self)
        self._init_toolbar()
        self.layout.addWidget(self.toolbar)

        # Placeholder for the actual ring progress bar widget
        self.ring_progress_bar = RingProgressContainerWidget(self)
        self.layout.addWidget(self.ring_progress_bar)

        self.settings_dialog = None

        self.toolbar.show_bundles(["rpb_settings"])

    def apply_theme(self, theme: str):
        super().apply_theme(theme)
        if self.ring_progress_bar.color_map:
            self.ring_progress_bar.set_colors_from_map(self.ring_progress_bar.color_map)

    def _init_toolbar(self):
        settings_action = MaterialIconAction(
            icon_name="settings",
            tooltip="Show Ring Progress Bar Settings",
            checkable=True,
            parent=self,
        )
        self.toolbar.add_action("rpb_settings", settings_action)
        settings_action.action.triggered.connect(self._open_settings_dialog)

    def _open_settings_dialog(self):
        """ "
        Open the settings dialog for the ring progress bar.
        """
        settings_action = self.toolbar.components.get_action("rpb_settings").action
        if self.settings_dialog is None or not self.settings_dialog.isVisible():
            settings = RingSettings(parent=self, target_widget=self, popup=True)
            self.settings_dialog = SettingsDialog(
                self,
                settings_widget=settings,
                window_title="Ring Progress Bar Settings",
                modal=False,
            )
            self.settings_dialog.resize(900, 500)
            self.settings_dialog.finished.connect(self._settings_dialog_closed)
            self.settings_dialog.show()

            settings_action.setChecked(True)
        else:
            # Dialog is already open, raise it
            self.settings_dialog.raise_()
            self.settings_dialog.activateWindow()
            settings_action.setChecked(True)

    def _settings_dialog_closed(self):
        """
        Handle the settings dialog being closed.
        """
        settings_action = self.toolbar.components.get_action("rpb_settings").action
        settings_action.setChecked(False)
        self.settings_dialog = None

    #################################################
    ###### RPC User Access Methods ##################
    #################################################

    def add_ring(self, config: dict | None = None) -> Ring:
        """
        Add a new ring to the ring progress bar.
        Optionally, a configuration dictionary can be provided but the ring
        can also be configured later. The config dictionary must provide
        the qproperties of the Qt Ring object.

        Args:
            config(dict | None): Optional configuration dictionary for the ring.

        Returns:
            Ring: The newly added ring object.
        """
        return self.ring_progress_bar.add_ring(config=config)

    def remove_ring(self, index: int | None = None):
        """
        Remove a ring from the ring progress bar.
        Args:
            index(int | None): Index of the ring to remove. If None, removes the last ring.
        """
        if self.ring_progress_bar.num_bars == 0:
            return
        self.ring_progress_bar.remove_ring(index=index)

    def set_gap(self, value: int):
        """
        Set the gap between rings.

        Args:
            value(int): Gap value in pixels.
        """
        self.gap = value

    def set_center_label(self, text: str):
        """
        Set the center label text.

        Args:
            text(str): Text for the center label.
        """
        self.center_label = text

    @property
    def rings(self) -> list[Ring]:
        return self.ring_progress_bar.rings

    ###############################################
    ####### QProperties ###########################
    ###############################################

    @SafeProperty(int)
    def gap(self) -> int:
        return self.ring_progress_bar.gap

    @gap.setter
    def gap(self, value: int):
        self.ring_progress_bar.gap = value
        self.ring_progress_bar.update()

    @SafeProperty(str)
    def color_map(self) -> str:
        return self.ring_progress_bar.color_map or ""

    @color_map.setter
    def color_map(self, colormap: str):
        if colormap == "":
            self.ring_progress_bar.color_map = ""
            return
        if colormap not in pg.colormap.listMaps():
            return
        self.ring_progress_bar.set_colors_from_map(colormap)
        self.ring_progress_bar.color_map = colormap

    @SafeProperty(str)
    def center_label(self) -> str:
        return self.ring_progress_bar.center_label.text()

    @center_label.setter
    def center_label(self, text: str):
        self.ring_progress_bar.center_label.setText(text)

    @SafeProperty(str, designable=False, popup_error=True)
    def ring_json(self) -> str:
        """
        A JSON string property that serializes all ring pydantic configs.
        """
        raw_list = []
        for ring in self.rings:
            cfg_dict = ring.config.model_dump()
            raw_list.append(cfg_dict)
        return json.dumps(raw_list, indent=2)

    @ring_json.setter
    def ring_json(self, json_data: str):
        """
        Load rings from a JSON string and add them to the ring progress bar.
        """
        try:
            ring_configs = json.loads(json_data)
            self.ring_progress_bar.clear_all()
            for cfg_dict in ring_configs:
                self.add_ring(config=cfg_dict)
        except json.JSONDecodeError as e:
            logger.error(f"Failed to decode JSON: {e}")

    def cleanup(self):
        self.ring_progress_bar.clear_all()
        self.ring_progress_bar.close()
        self.ring_progress_bar.deleteLater()
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.utils.colors import apply_theme

    app = QApplication(sys.argv)
    apply_theme("dark")
    widget = RingProgressBar()
    widget.show()
    sys.exit(app.exec_())



from bec_lib.device import ReadoutPriority
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QStyledItemDelegate, QWidget

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.toolbars.actions import DeviceComboBoxAction, WidgetAction
from bec_widgets.utils.toolbars.bundles import ToolbarComponents
from bec_widgets.utils.toolbars.toolbar import ToolbarBundle
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import BECDeviceFilter
from bec_widgets.widgets.control.device_input.device_combobox.device_combobox import DeviceComboBox
from bec_widgets.widgets.utility.visual.colormap_widget.colormap_widget import BECColorMapWidget


class NoCheckDelegate(QStyledItemDelegate):
    """To reduce space in combo boxes by removing the checkmark."""

    def initStyleOption(self, option, index):
        super().initStyleOption(option, index)
        # Remove any check indicator
        option.checkState = Qt.Unchecked


def monitor_selection_bundle(
    components: ToolbarComponents, target_widget: QWidget
) -> ToolbarBundle:
    """
    Creates a monitor selection toolbar bundle.

    Args:
        components (ToolbarComponents): The components to be added to the bundle.

    Returns:
        ToolbarBundle: The monitor selection toolbar bundle.
    """
    components.add_safe(
        "monitor_selection",
        DeviceComboBoxAction(
            target_widget=target_widget,
            device_filter=[BECDeviceFilter.DEVICE],
            readout_priority_filter=ReadoutPriority.ASYNC,
            add_empty_item=True,
            no_check_delegate=True,
        ),
    )
    components.add_safe(
        "color_map", WidgetAction(widget=BECColorMapWidget(cmap="plasma"), adjust_size=False)
    )
    bundle = ToolbarBundle("monitor_selection", components)
    bundle.add_action("monitor_selection")
    bundle.add_action("color_map")
    return bundle


class MultiWaveformSelectionToolbarBundle(ToolbarBundle):
    """
    A bundle of actions for a toolbar that selects motors.
    """

    def __init__(self, bundle_id="monitor_selection", target_widget=None, **kwargs):
        super().__init__(bundle_id=bundle_id, actions=[], **kwargs)
        self.target_widget = target_widget

        # Monitor Selection
        self.monitor = DeviceComboBox(
            device_filter=BECDeviceFilter.DEVICE, readout_priority_filter=ReadoutPriority.ASYNC
        )
        self.monitor.addItem("", None)
        self.monitor.setCurrentText("")
        self.monitor.setToolTip("Select Monitor")
        self.monitor.setItemDelegate(NoCheckDelegate(self.monitor))
        self.add_action("monitor", WidgetAction(widget=self.monitor, adjust_size=False))

        # Colormap Selection
        self.colormap_widget = BECColorMapWidget(cmap="plasma")
        self.add_action("color_map", WidgetAction(widget=self.colormap_widget, adjust_size=False))

        # Connect slots, a device will be connected upon change of any combobox
        self.monitor.currentTextChanged.connect(lambda: self.connect())
        self.colormap_widget.colormap_changed_signal.connect(self.change_colormap)

    @SafeSlot()
    def connect(self):
        monitor = self.monitor.currentText()

        if monitor != "":
            if monitor != self.target_widget.config.monitor:
                self.target_widget.monitor = monitor

    @SafeSlot(str)
    def change_colormap(self, colormap: str):
        self.target_widget.color_palette = colormap

    def cleanup(self):
        """
        Cleanup the toolbar bundle.
        """
        self.monitor.close()
        self.monitor.deleteLater()



from __future__ import annotations

from qtpy.QtCore import QLocale, QMetaEnum, Qt, QTimer
from qtpy.QtGui import QColor, QCursor, QFont, QIcon, QPalette
from qtpy.QtWidgets import (
    QCheckBox,
    QColorDialog,
    QComboBox,
    QDoubleSpinBox,
    QFileDialog,
    QFontDialog,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QMenu,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QToolButton,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)


class PropertyEditor(QWidget):
    def __init__(self, target: QWidget, parent: QWidget | None = None, show_only_bec: bool = True):
        super().__init__(parent)
        self._target = target
        self._bec_only = show_only_bec

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Name row
        name_row = QHBoxLayout()
        name_row.addWidget(QLabel("Name:"))
        self.name_edit = QLineEdit(target.objectName())
        self.name_edit.setEnabled(False)  # TODO implement with RPC broadcast
        name_row.addWidget(self.name_edit)
        layout.addLayout(name_row)

        # BEC only checkbox
        filter_row = QHBoxLayout()
        self.chk_show_qt = QCheckBox("Show Qt properties")
        self.chk_show_qt.setChecked(False)
        filter_row.addWidget(self.chk_show_qt)
        filter_row.addStretch(1)
        layout.addLayout(filter_row)
        self.chk_show_qt.toggled.connect(lambda checked: self.set_show_only_bec(not checked))

        # Main tree widget
        self.tree = QTreeWidget(self)
        self.tree.setColumnCount(2)
        self.tree.setHeaderLabels(["Property", "Value"])
        self.tree.setAlternatingRowColors(True)
        self.tree.setRootIsDecorated(False)
        layout.addWidget(self.tree)
        self._build()

    def _class_chain(self):
        chain = []
        mo = self._target.metaObject()
        while mo is not None:
            chain.append(mo)
            mo = mo.superClass()
        return chain

    def set_show_only_bec(self, flag: bool):
        self._bec_only = flag
        self._build()

    def _set_equal_columns(self):
        header = self.tree.header()
        header.setSectionResizeMode(0, QHeaderView.Interactive)
        header.setSectionResizeMode(1, QHeaderView.Interactive)
        w = self.tree.viewport().width() or self.tree.width()
        if w > 0:
            half = max(1, w // 2)
            self.tree.setColumnWidth(0, half)
            self.tree.setColumnWidth(1, w - half)

    def _build(self):
        self.tree.clear()
        for mo in self._class_chain():
            class_name = mo.className()
            if self._bec_only and not self._is_bec_metaobject(mo):
                continue
            group_item = QTreeWidgetItem(self.tree, [class_name])
            group_item.setFirstColumnSpanned(True)
            start = mo.propertyOffset()
            end = mo.propertyCount()
            for i in range(start, end):
                prop = mo.property(i)
                if (
                    not prop.isReadable()
                    or not prop.isWritable()
                    or not prop.isStored()
                    or not prop.isDesignable()
                ):
                    continue
                name = prop.name()
                if name == "objectName":
                    continue
                value = self._target.property(name)
                self._add_property_row(group_item, name, value, prop)
            if group_item.childCount() == 0:
                idx = self.tree.indexOfTopLevelItem(group_item)
                self.tree.takeTopLevelItem(idx)
        self.tree.expandAll()
        QTimer.singleShot(0, self._set_equal_columns)

    def _enum_int(self, obj) -> int:
        return int(getattr(obj, "value", obj))

    def _make_sizepolicy_editor(self, name: str, sp):
        if not isinstance(sp, QSizePolicy):
            return None
        wrap = QWidget(self)
        row = QHBoxLayout(wrap)
        row.setContentsMargins(0, 0, 0, 0)
        row.setSpacing(4)
        h_combo = QComboBox(wrap)
        v_combo = QComboBox(wrap)
        hs = QSpinBox(wrap)
        vs = QSpinBox(wrap)
        for b in (hs, vs):
            b.setRange(0, 16777215)
        policies = [
            (QSizePolicy.Fixed, "Fixed"),
            (QSizePolicy.Minimum, "Minimum"),
            (QSizePolicy.Maximum, "Maximum"),
            (QSizePolicy.Preferred, "Preferred"),
            (QSizePolicy.Expanding, "Expanding"),
            (QSizePolicy.MinimumExpanding, "MinExpanding"),
            (QSizePolicy.Ignored, "Ignored"),
        ]
        for pol, text in policies:
            h_combo.addItem(text, self._enum_int(pol))
            v_combo.addItem(text, self._enum_int(pol))

        def _set_current(combo, val):
            idx = combo.findData(self._enum_int(val))
            if idx >= 0:
                combo.setCurrentIndex(idx)

        _set_current(h_combo, sp.horizontalPolicy())
        _set_current(v_combo, sp.verticalPolicy())
        hs.setValue(sp.horizontalStretch())
        vs.setValue(sp.verticalStretch())

        def apply_changes():
            hp = QSizePolicy.Policy(h_combo.currentData())
            vp = QSizePolicy.Policy(v_combo.currentData())
            nsp = QSizePolicy(hp, vp)
            nsp.setHorizontalStretch(hs.value())
            nsp.setVerticalStretch(vs.value())
            self._target.setProperty(name, nsp)

        h_combo.currentIndexChanged.connect(lambda _=None: apply_changes())
        v_combo.currentIndexChanged.connect(lambda _=None: apply_changes())
        hs.valueChanged.connect(lambda _=None: apply_changes())
        vs.valueChanged.connect(lambda _=None: apply_changes())
        row.addWidget(h_combo)
        row.addWidget(v_combo)
        row.addWidget(hs)
        row.addWidget(vs)
        return wrap

    def _make_locale_editor(self, name: str, loc):
        if not isinstance(loc, QLocale):
            return None
        wrap = QWidget(self)
        row = QHBoxLayout(wrap)
        row.setContentsMargins(0, 0, 0, 0)
        row.setSpacing(4)
        lang_combo = QComboBox(wrap)
        country_combo = QComboBox(wrap)
        for lang in QLocale.Language:
            try:
                lang_int = self._enum_int(lang)
            except Exception:
                continue
            if lang_int < 0:
                continue
            name_txt = QLocale.languageToString(QLocale.Language(lang_int))
            lang_combo.addItem(name_txt, lang_int)

        def populate_countries():
            country_combo.blockSignals(True)
            country_combo.clear()
            for terr in QLocale.Country:
                try:
                    terr_int = self._enum_int(terr)
                except Exception:
                    continue
                if terr_int < 0:
                    continue
                text = QLocale.countryToString(QLocale.Country(terr_int))
                country_combo.addItem(text, terr_int)
            cur_country = self._enum_int(loc.country())
            idx = country_combo.findData(cur_country)
            if idx >= 0:
                country_combo.setCurrentIndex(idx)
            country_combo.blockSignals(False)

        cur_lang = self._enum_int(loc.language())
        idx = lang_combo.findData(cur_lang)
        if idx >= 0:
            lang_combo.setCurrentIndex(idx)
        populate_countries()

        def apply_locale():
            lang = QLocale.Language(int(lang_combo.currentData()))
            country = QLocale.Country(int(country_combo.currentData()))
            self._target.setProperty(name, QLocale(lang, country))

        lang_combo.currentIndexChanged.connect(lambda _=None: populate_countries())
        lang_combo.currentIndexChanged.connect(lambda _=None: apply_locale())
        country_combo.currentIndexChanged.connect(lambda _=None: apply_locale())
        row.addWidget(lang_combo)
        row.addWidget(country_combo)
        return wrap

    def _make_icon_editor(self, name: str, icon):
        btn = QPushButton(self)
        btn.setText("Choose…")
        if isinstance(icon, QIcon) and not icon.isNull():
            btn.setIcon(icon)

        def pick():
            path, _ = QFileDialog.getOpenFileName(
                self, "Select Icon", "", "Images (*.png *.jpg *.jpeg *.bmp *.svg)"
            )
            if path:
                ic = QIcon(path)
                self._target.setProperty(name, ic)
                btn.setIcon(ic)

        btn.clicked.connect(pick)
        return btn

    def _spin_pair(self, ints: bool = True):
        box1 = QSpinBox(self) if ints else QDoubleSpinBox(self)
        box2 = QSpinBox(self) if ints else QDoubleSpinBox(self)
        if ints:
            box1.setRange(-10_000_000, 10_000_000)
            box2.setRange(-10_000_000, 10_000_000)
        else:
            for b in (box1, box2):
                b.setDecimals(6)
                b.setRange(-1e12, 1e12)
                b.setSingleStep(0.1)
        row = QHBoxLayout()
        row.setContentsMargins(0, 0, 0, 0)
        row.setSpacing(4)
        wrap = QWidget(self)
        wrap.setLayout(row)
        row.addWidget(box1)
        row.addWidget(box2)
        return wrap, box1, box2

    def _spin_quad(self, ints: bool = True):
        s = QSpinBox if ints else QDoubleSpinBox
        boxes = [s(self) for _ in range(4)]
        if ints:
            for b in boxes:
                b.setRange(-10_000_000, 10_000_000)
        else:
            for b in boxes:
                b.setDecimals(6)
                b.setRange(-1e12, 1e12)
                b.setSingleStep(0.1)
        row = QHBoxLayout()
        row.setContentsMargins(0, 0, 0, 0)
        row.setSpacing(4)
        wrap = QWidget(self)
        wrap.setLayout(row)
        for b in boxes:
            row.addWidget(b)
        return wrap, boxes

    def _make_font_editor(self, name: str, value):
        btn = QPushButton(self)
        if isinstance(value, QFont):
            btn.setText(f"{value.family()}, {value.pointSize()}pt")
        else:
            btn.setText("Select font…")

        def pick():
            ok, font = QFontDialog.getFont(
                value if isinstance(value, QFont) else QFont(), self, "Select Font"
            )
            if ok:
                self._target.setProperty(name, font)
                btn.setText(f"{font.family()}, {font.pointSize()}pt")

        btn.clicked.connect(pick)
        return btn

    def _make_color_editor(self, initial: QColor, apply_cb):
        btn = QPushButton(self)
        if isinstance(initial, QColor):
            btn.setText(initial.name())
            btn.setStyleSheet(f"background:{initial.name()};")
        else:
            btn.setText("Select color…")

        def pick():
            col = QColorDialog.getColor(
                initial if isinstance(initial, QColor) else QColor(), self, "Select Color"
            )
            if col.isValid():
                apply_cb(col)
                btn.setText(col.name())
                btn.setStyleSheet(f"background:{col.name()};")

        btn.clicked.connect(pick)
        return btn

    def _apply_palette_color(
        self,
        name: str,
        pal: QPalette,
        group: QPalette.ColorGroup,
        role: QPalette.ColorRole,
        col: QColor,
    ):
        pal.setColor(group, role, col)
        self._target.setProperty(name, pal)

    def _make_palette_editor(self, name: str, pal: QPalette):
        if not isinstance(pal, QPalette):
            return None
        wrap = QWidget(self)
        row = QHBoxLayout(wrap)
        row.setContentsMargins(0, 0, 0, 0)
        group_combo = QComboBox(wrap)
        role_combo = QComboBox(wrap)
        pick_btn = self._make_color_editor(
            pal.color(QPalette.Active, QPalette.WindowText),
            lambda col: self._apply_palette_color(
                name, pal, QPalette.Active, QPalette.WindowText, col
            ),
        )
        groups = [
            (QPalette.Active, "Active"),
            (QPalette.Inactive, "Inactive"),
            (QPalette.Disabled, "Disabled"),
        ]
        for g, label in groups:
            group_combo.addItem(label, int(getattr(g, "value", g)))
        roles = [
            (QPalette.WindowText, "WindowText"),
            (QPalette.Window, "Window"),
            (QPalette.Base, "Base"),
            (QPalette.AlternateBase, "AlternateBase"),
            (QPalette.ToolTipBase, "ToolTipBase"),
            (QPalette.ToolTipText, "ToolTipText"),
            (QPalette.Text, "Text"),
            (QPalette.Button, "Button"),
            (QPalette.ButtonText, "ButtonText"),
            (QPalette.BrightText, "BrightText"),
            (QPalette.Highlight, "Highlight"),
            (QPalette.HighlightedText, "HighlightedText"),
        ]
        for r, label in roles:
            role_combo.addItem(label, int(getattr(r, "value", r)))

        def rewire_button():
            g = QPalette.ColorGroup(int(group_combo.currentData()))
            r = QPalette.ColorRole(int(role_combo.currentData()))
            col = pal.color(g, r)
            while row.count() > 2:
                w = row.takeAt(2).widget()
                if w:
                    w.deleteLater()
            btn = self._make_color_editor(
                col, lambda c: self._apply_palette_color(name, pal, g, r, c)
            )
            row.addWidget(btn)

        group_combo.currentIndexChanged.connect(lambda _: rewire_button())
        role_combo.currentIndexChanged.connect(lambda _: rewire_button())
        row.addWidget(group_combo)
        row.addWidget(role_combo)
        row.addWidget(pick_btn)
        return wrap

    def _make_cursor_editor(self, name: str, value):
        combo = QComboBox(self)
        shapes = [
            (Qt.ArrowCursor, "Arrow"),
            (Qt.IBeamCursor, "IBeam"),
            (Qt.WaitCursor, "Wait"),
            (Qt.CrossCursor, "Cross"),
            (Qt.UpArrowCursor, "UpArrow"),
            (Qt.SizeAllCursor, "SizeAll"),
            (Qt.PointingHandCursor, "PointingHand"),
            (Qt.ForbiddenCursor, "Forbidden"),
            (Qt.WhatsThisCursor, "WhatsThis"),
            (Qt.BusyCursor, "Busy"),
        ]
        current_shape = None
        if isinstance(value, QCursor):
            try:
                enum_val = value.shape()
                current_shape = int(getattr(enum_val, "value", enum_val))
            except Exception:
                current_shape = None
        for shape, text in shapes:
            combo.addItem(text, int(getattr(shape, "value", shape)))
        if current_shape is not None:
            idx = combo.findData(current_shape)
            if idx >= 0:
                combo.setCurrentIndex(idx)

        def apply_index(i):
            shape_val = int(combo.itemData(i))
            self._target.setProperty(name, QCursor(Qt.CursorShape(shape_val)))

        combo.currentIndexChanged.connect(apply_index)
        return combo

    def _add_property_row(self, parent: QTreeWidgetItem, name: str, value, prop):
        item = QTreeWidgetItem(parent, [name, ""])
        editor = self._make_editor(name, value, prop)
        if editor is not None:
            self.tree.setItemWidget(item, 1, editor)
        else:
            item.setText(1, repr(value))

    def _is_bec_metaobject(self, mo) -> bool:
        cname = mo.className()
        for cls in type(self._target).mro():
            if getattr(cls, "__name__", None) == cname:
                mod = getattr(cls, "__module__", "")
                return mod.startswith("bec_widgets")
        return False

    def _enum_text(self, meta_enum: QMetaEnum, value_int: int) -> str:
        if not meta_enum.isFlag():
            key = meta_enum.valueToKey(value_int)
            return key.decode() if isinstance(key, (bytes, bytearray)) else (key or str(value_int))
        parts = []
        for i in range(meta_enum.keyCount()):
            k = meta_enum.key(i)
            v = meta_enum.value(i)
            if value_int & v:
                k = k.decode() if isinstance(k, (bytes, bytearray)) else k
                parts.append(k)
        return " | ".join(parts) if parts else "0"

    def _enum_value_to_int(self, meta_enum: QMetaEnum, value) -> int:
        try:
            return int(value)
        except Exception:
            pass
        v = getattr(value, "value", None)
        if isinstance(v, (int,)):
            return int(v)
        n = getattr(value, "name", None)
        if isinstance(n, str):
            res = meta_enum.keyToValue(n)
            if res != -1:
                return int(res)
        s = str(value)
        parts = [p.strip() for p in s.replace(",", "|").split("|")]
        keys = []
        for p in parts:
            if "." in p:
                p = p.split(".")[-1]
            keys.append(p)
        keystr = "|".join(keys)
        try:
            res = meta_enum.keysToValue(keystr)
            if res != -1:
                return int(res)
        except Exception:
            pass
        return 0

    def _make_enum_editor(self, name: str, value, prop):
        meta_enum = prop.enumerator()
        current = self._enum_value_to_int(meta_enum, value)

        if not meta_enum.isFlag():
            combo = QComboBox(self)
            for i in range(meta_enum.keyCount()):
                key = meta_enum.key(i)
                key = key.decode() if isinstance(key, (bytes, bytearray)) else key
                combo.addItem(key, meta_enum.value(i))
            idx = combo.findData(current)
            if idx < 0:
                txt = self._enum_text(meta_enum, current)
                idx = combo.findText(txt)
            combo.setCurrentIndex(max(idx, 0))

            def apply_index(i):
                v = combo.itemData(i)
                self._target.setProperty(name, int(v))

            combo.currentIndexChanged.connect(apply_index)
            return combo

        btn = QToolButton(self)
        btn.setText(self._enum_text(meta_enum, current))
        btn.setPopupMode(QToolButton.InstantPopup)
        menu = QMenu(btn)
        actions = []
        for i in range(meta_enum.keyCount()):
            key = meta_enum.key(i)
            key = key.decode() if isinstance(key, (bytes, bytearray)) else key
            act = menu.addAction(key)
            act.setCheckable(True)
            act.setChecked(bool(current & meta_enum.value(i)))
            actions.append(act)
        btn.setMenu(menu)

        def apply_flags():
            flags = 0
            for i, act in enumerate(actions):
                if act.isChecked():
                    flags |= meta_enum.value(i)
            self._target.setProperty(name, int(flags))
            btn.setText(self._enum_text(meta_enum, flags))

        menu.triggered.connect(lambda _a: apply_flags())
        return btn

    def _make_editor(self, name: str, value, prop):
        from qtpy.QtCore import QPoint, QPointF, QRect, QRectF, QSize, QSizeF

        if prop.isEnumType():
            return self._make_enum_editor(name, value, prop)
        if isinstance(value, QColor):
            return self._make_color_editor(value, lambda col: self._target.setProperty(name, col))
        if isinstance(value, QFont):
            return self._make_font_editor(name, value)
        if isinstance(value, QPalette):
            return self._make_palette_editor(name, value)
        if isinstance(value, QCursor):
            return self._make_cursor_editor(name, value)
        if isinstance(value, QSizePolicy):
            ed = self._make_sizepolicy_editor(name, value)
            if ed is not None:
                return ed
        if isinstance(value, QLocale):
            ed = self._make_locale_editor(name, value)
            if ed is not None:
                return ed
        if isinstance(value, QIcon):
            ed = self._make_icon_editor(name, value)
            if ed is not None:
                return ed
        if isinstance(value, QSize):
            wrap, w, h = self._spin_pair(ints=True)
            w.setValue(value.width())
            h.setValue(value.height())
            w.valueChanged.connect(
                lambda _: self._target.setProperty(name, QSize(w.value(), h.value()))
            )
            h.valueChanged.connect(
                lambda _: self._target.setProperty(name, QSize(w.value(), h.value()))
            )
            return wrap
        if isinstance(value, QSizeF):
            wrap, w, h = self._spin_pair(ints=False)
            w.setValue(value.width())
            h.setValue(value.height())
            w.valueChanged.connect(
                lambda _: self._target.setProperty(name, QSizeF(w.value(), h.value()))
            )
            h.valueChanged.connect(
                lambda _: self._target.setProperty(name, QSizeF(w.value(), h.value()))
            )
            return wrap
        if isinstance(value, QPoint):
            wrap, x, y = self._spin_pair(ints=True)
            x.setValue(value.x())
            y.setValue(value.y())
            x.valueChanged.connect(
                lambda _: self._target.setProperty(name, QPoint(x.value(), y.value()))
            )
            y.valueChanged.connect(
                lambda _: self._target.setProperty(name, QPoint(x.value(), y.value()))
            )
            return wrap
        if isinstance(value, QPointF):
            wrap, x, y = self._spin_pair(ints=False)
            x.setValue(value.x())
            y.setValue(value.y())
            x.valueChanged.connect(
                lambda _: self._target.setProperty(name, QPointF(x.value(), y.value()))
            )
            y.valueChanged.connect(
                lambda _: self._target.setProperty(name, QPointF(x.value(), y.value()))
            )
            return wrap
        if isinstance(value, QRect):
            wrap, boxes = self._spin_quad(ints=True)
            for b, v in zip(boxes, (value.x(), value.y(), value.width(), value.height())):
                b.setValue(v)

            def apply_rect():
                self._target.setProperty(
                    name,
                    QRect(boxes[0].value(), boxes[1].value(), boxes[2].value(), boxes[3].value()),
                )

            for b in boxes:
                b.valueChanged.connect(lambda _=None: apply_rect())
            return wrap
        if isinstance(value, QRectF):
            wrap, boxes = self._spin_quad(ints=False)
            for b, v in zip(boxes, (value.x(), value.y(), value.width(), value.height())):
                b.setValue(v)

            def apply_rectf():
                self._target.setProperty(
                    name,
                    QRectF(boxes[0].value(), boxes[1].value(), boxes[2].value(), boxes[3].value()),
                )

            for b in boxes:
                b.valueChanged.connect(lambda _=None: apply_rectf())
            return wrap
        if isinstance(value, bool):
            w = QCheckBox(self)
            w.setChecked(bool(value))
            w.toggled.connect(lambda v: self._target.setProperty(name, v))
            return w
        if isinstance(value, int) and not isinstance(value, bool):
            w = QSpinBox(self)
            w.setRange(-10_000_000, 10_000_000)
            w.setValue(int(value))
            w.valueChanged.connect(lambda v: self._target.setProperty(name, v))
            return w
        if isinstance(value, float):
            w = QDoubleSpinBox(self)
            w.setDecimals(6)
            w.setRange(-1e12, 1e12)
            w.setSingleStep(0.1)
            w.setValue(float(value))
            w.valueChanged.connect(lambda v: self._target.setProperty(name, v))
            return w
        if isinstance(value, str):
            w = QLineEdit(self)
            w.setText(value)
            w.editingFinished.connect(lambda: self._target.setProperty(name, w.text()))
            return w
        return None


class DemoApp(QWidget):  # pragma: no cover:
    def __init__(self):
        super().__init__()

        layout = QHBoxLayout(self)

        # Create a BECWidget instance example
        waveform = self.create_waveform()

        # property editor for the BECWidget
        property_editor = PropertyEditor(waveform, show_only_bec=True)

        layout.addWidget(waveform)
        layout.addWidget(property_editor)

    def create_waveform(self):
        """Create a new waveform widget."""

        from bec_widgets.widgets.plots.waveform.waveform import Waveform

        waveform = Waveform(parent=self)
        waveform.title = "New Waveform"
        waveform.x_label = "X Axis"
        waveform.y_label = "Y Axis"
        return waveform


if __name__ == "__main__":  # pragma: no cover:
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    demo = DemoApp()
    demo.setWindowTitle("Property Editor Demo")
    demo.resize(1200, 800)
    demo.show()
    sys.exit(app.exec())
